/* $Id: cryptoapi.c,v 1.5 2007/02/02 06:32:42 ezk Exp $
 *
 * crypto/cryptoapi.c
 *
 * Written by Alexander Kjeldaas <astor@fast.no> 1998-11-15
 *
 * 2000-10-15 Harald Welte <laforge@gnumonks.org>
 * 		- ported to Linux 2.4
 *
 * Copyright (C)  1998 Alexander Kjeldaas.
 *                2001 Herbert Valerio Riedel
 *
 * This module is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/version.h>

#include <linux/init.h>
#include <linux/config.h>
#include <linux/string.h>
#include "crypto.h"

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif /* CONFIG_KMOD */

static struct tf_cipher_context *
default_realloc_tf_cipher_context(struct tf_cipher_context *old_cx,
			       struct tf_cipher_implementation *,
			       int max_key_len);

static void default_wipe_context(struct tf_cipher_context *cx);

static void default_free_tf_cipher_context(struct tf_cipher_context *cx);
/* ciphers */
static int default_encrypt(struct tf_cipher_context *cx,
			   const u8 *in, u8 *out, int size);
static int default_encrypt_atomic(struct tf_cipher_context *cx,
				  const u8 *in, u8 *out, int size);
static int default_encrypt_iv(struct tf_cipher_context *cx,
			      const u8 *in, u8 *out, int size,
			      const u8 *iv);
static int default_encrypt_atomic_iv(struct tf_cipher_context *cx,
				     const u8 *in, u8 *out, int size,
				     const u8 *iv);

static int default_decrypt(struct tf_cipher_context *cx,
			   const u8 *in, u8 *out, int size);
static int default_decrypt_atomic(struct tf_cipher_context *cx,
				  const u8 *in, u8 *out, int size);

static int default_decrypt_iv(struct tf_cipher_context *cx,
			      const u8 *in, u8 *out, int size, const u8 *iv);
static int default_decrypt_atomic_iv(struct tf_cipher_context *cx,
				     const u8 *in, u8 *out, int size,
				     const u8 *iv);

static int default_set_key(struct tf_cipher_context *cx,
			   const u8 *key, int key_len);
static int default_set_key_atomic(struct tf_cipher_context *cx,
				  const u8 *key, int key_len);

int
tf_register_cipher(struct tf_cipher_implementation *ci)
{
	if (!ci->realloc_context)
		ci->realloc_context = default_realloc_tf_cipher_context;

	if (!ci->wipe_context)
		ci->wipe_context = default_wipe_context;

	if (!ci->free_context)
		ci->free_context = default_free_tf_cipher_context;

	if (ci->_encrypt && ci->_decrypt && ci->_set_key) {
		ci->encrypt_iv = default_encrypt_iv;
		ci->encrypt_atomic_iv = default_encrypt_atomic_iv;
		ci->decrypt_iv = default_decrypt_iv;
		ci->decrypt_atomic_iv = default_decrypt_atomic_iv;

		ci->encrypt = default_encrypt;
		ci->encrypt_atomic = default_encrypt_atomic;
		ci->decrypt = default_decrypt;
		ci->decrypt_atomic = default_decrypt_atomic;
		ci->set_key = default_set_key;
		ci->set_key_atomic = default_set_key_atomic;
	}

	if (!ci->encrypt || !ci->decrypt || !ci->set_key) {
		return -EINVAL;
	}
	return 0;
}

int
tf_unregister_cipher(struct tf_cipher_implementation *ci)
{
	// nothing
	return 0;
}

struct tf_cipher_context *
default_realloc_tf_cipher_context(struct tf_cipher_context *old_cx,
			       struct tf_cipher_implementation *ci,
			       int max_key_len)
{
	struct tf_cipher_context *cx;
	/* Default ciphers need the same amount of memory for any key
           size */
	if (old_cx) {
		return old_cx;
	}
	cx = kmalloc(sizeof(struct tf_cipher_context) +
		     ci->key_schedule_size, GFP_KERNEL);
	if (!cx) {
		return NULL;
	}
	cx->ci = ci;
	cx->keyinfo = (void *)((char *)cx)+sizeof(struct tf_cipher_context);
	(void) max_key_len; /* Make gcc happy */
	return cx;
}

void
default_wipe_context(struct tf_cipher_context *cx)
{
	struct tf_cipher_implementation *ci = cx->ci;
	u32 *keyinfo = cx->keyinfo;
	memset(cx->keyinfo, 0, ci->key_schedule_size);
	memset(cx, 0, sizeof(struct tf_cipher_context));
	cx->ci = ci;
	cx->keyinfo = keyinfo;
}

void
default_free_tf_cipher_context(struct tf_cipher_context *cx)
{
	kfree(cx);
}

static int
default_encrypt(struct tf_cipher_context *cx, const u8 *in, u8 *out, int size)
{
	return cx->ci->_encrypt(cx, in, out, size, 0, cx->iv);
}

static int
default_encrypt_atomic(struct tf_cipher_context *cx, const u8 *in, u8 *out,
		       int size)
{
	return cx->ci->_encrypt(cx, in, out, size, 1, cx->iv);
}

static int
default_encrypt_iv(struct tf_cipher_context *cx, const u8 *in, u8 *out, int size,
		const u8 *iv)
{
	return cx->ci->_encrypt(cx, in, out, size, 0, iv);
}

static int
default_encrypt_atomic_iv(struct tf_cipher_context *cx, const u8 *in, u8 *out,
		          int size, const u8 *iv)
{
	return cx->ci->_encrypt(cx, in, out, size, 1, iv);
}

static int
default_decrypt(struct tf_cipher_context *cx, const u8 *in, u8 *out, int size)
{
	return cx->ci->_decrypt(cx, in, out, size, 0, cx->iv);
}

static int
default_decrypt_atomic(struct tf_cipher_context *cx, const u8 *in, u8 *out,
		       int size)
{
	return cx->ci->_decrypt(cx, in, out, size, 1, cx->iv);
}

static int
default_decrypt_iv(struct tf_cipher_context *cx, const u8 *in, u8 *out, int size,
		const u8 *iv)
{
	return cx->ci->_decrypt(cx, in, out, size, 0, iv);
}

static int
default_decrypt_atomic_iv(struct tf_cipher_context *cx, const u8 *in, u8 *out,
		       int size, const u8 *iv)
{
	return cx->ci->_decrypt(cx, in, out, size, 1, iv);
}


static int
default_set_key(struct tf_cipher_context *cx, const u8 *key, int key_len)
{
	return cx->ci->_set_key(cx, key, key_len, 0);
}

static int
default_set_key_atomic(struct tf_cipher_context *cx, const u8 *key,
		       int key_len)
{
	return cx->ci->_set_key(cx, key, key_len, 1);
}
