/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 * Copyright (c) 2004-2005 Timothy Wong
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */
/*
 *  $Id: main.c,v 1.26 2007/01/30 06:30:53 gopalas Exp $
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif /* HAVE_CONFIG_H */
#ifdef FISTGEN
# include "fist_tracefs.h"
#endif /* FISTGEN */
#include "fist.h"
#include "tracefs.h"
#include "tracer.h"
#include <linux/module.h>
#ifdef KWRAPPERS
#include "kwrappers.h"
#endif
/* This definition must only appear after we include <linux/module.h> */
#ifndef MODULE_LICENSE
#define MODULE_LICENSE(foo)
#endif /* not MODULE_LICENSE */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
#define get_inode_idev(inode) inode->i_dev
#else
#include <linux/namei.h>
#define get_inode_idev(inode) inode->i_sb->s_dev
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

extern int fist_debug_var;

/* sb we pass is tracefs's super_block */
int
tracefs_interpose(dentry_t *hidden_dentry, dentry_t *dentry, super_block_t *sb, int flag)
{
	inode_t *hidden_inode;
	int err = 0;
	inode_t *inode;

	print_entry_location();

	hidden_inode = hidden_dentry->d_inode; /* CPW: moved after print_entry_location */

	ASSERT(hidden_inode != NULL);
	ASSERT(dentry->d_inode == NULL);

	/*
	 * We allocate our new inode below, by calling iget.
	 * iget will call our read_inode which will initialize some
	 * of the new inode's fields
	 */
	/* check that the lower file system didn't cross a mount point */
	if (hidden_inode->i_sb != stohs(sb)) {
		err = -EXDEV;
		goto out;
	}
	inode = iget(sb, hidden_inode->i_ino);

	if (!inode) {
		err = -EACCES;		/* should be impossible??? */
		goto out;
	}

	/*
	 * interpose the inode if not already interposed
	 *   this is possible if the inode is being reused
	 * XXX: what happens if we get_empty_inode() but there's another already?
	 * for now, ASSERT() that this can't happen; fix later.
	 */
	if (itohi(inode) == NULL)
		itohi(inode) = igrab(hidden_inode);

	/* Use different set of inode ops for symlinks & directories*/
	if (S_ISLNK(hidden_inode->i_mode))
		inode->i_op = &tracefs_symlink_iops;
	else if (S_ISDIR(hidden_inode->i_mode))
		inode->i_op = &tracefs_dir_iops;
	/* Use different set of file ops for directories */
	if (S_ISDIR(hidden_inode->i_mode))
		inode->i_fop = &tracefs_dir_fops;

	/* properly initialize special inodes */
	if (S_ISBLK(hidden_inode->i_mode) || S_ISCHR(hidden_inode->i_mode) ||
	    S_ISFIFO(hidden_inode->i_mode) || S_ISSOCK(hidden_inode->i_mode)) {
		init_special_inode(inode, hidden_inode->i_mode, hidden_inode->i_rdev);
	}

	/* Fix our inode's address operations to that of the lower inode */

	if (!mmap_support(inode->i_sb) && (inode->i_mapping->a_ops != hidden_inode->i_mapping->a_ops)) {
		fist_dprint(7, "fixing inode 0x%x a_ops (0x%x -> 0x%x)\n",
				(int) inode, (int) inode->i_mapping->a_ops,
				(int) hidden_inode->i_mapping->a_ops);
		inode->i_mapping->a_ops = hidden_inode->i_mapping->a_ops;
	}

	/* only (our) lookup wants to do a d_add */
	if (flag)
		d_add(dentry, inode);
	else
		d_instantiate(dentry, inode);

	ASSERT(dtopd(dentry) != NULL);

	/* all well, copy inode attributes */
	fist_copy_attr_all(inode, hidden_inode);

out:
	print_exit_status(err);
	return err;
}


/* find hidden dentry given this tracefs dentry */
dentry_t *
__tracefs_hidden_dentry(const char *file, const char *func, int line, dentry_t *dentry)
{
    dentry_t *hidden_dentry;

    ASSERT2(dentry != NULL);
    ASSERT2(dentry->d_op != NULL);
    ASSERT2(dentry->d_op == &tracefs_dops
	    );
    ASSERT2(dentry->d_sb->s_op == &tracefs_sops);
    if (dentry->d_inode)
	ASSERT2(dentry->d_inode->i_op == &tracefs_main_iops ||
		dentry->d_inode->i_op == &tracefs_dir_iops ||
		dentry->d_inode->i_op == &tracefs_symlink_iops);

    hidden_dentry = dtohd(dentry);
    ASSERT2(hidden_dentry != NULL);
    return hidden_dentry;
}


/*
 * Parse mount options: dir=XXX and debug=N
 *
 * Returns the dentry object of the lower-level (hidden) directory;
 * We want to mount our stackable file system on top of that hidden directory.
 *
 * Sets default debugging level to N, if any.
 */
dentry_t *
tracefs_parse_options(super_block_t *sb, char *options)
{
    dentry_t *hidden_root = ERR_PTR(-EINVAL);
    struct nameidata nd;
    char *name, *tmp, *end;
    int err = 0;

    print_entry_location();

    /* We don't want to go off the end of our arguments later on. */
    for (end = options; *end; end++);

    while (options < end) {
	tmp = options;
	while (*tmp && *tmp != ',')
	    tmp++;
	*tmp = '\0';
	if (!strncmp("dir=", options, 4)) {
	    /* note: the name passed need not be encoded */
	    name = options + 4;
	    fist_dprint(4, "tracefs: using directory: %s\n", name);
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    if (path_init(name, LOOKUP_FOLLOW, &nd))
		err = path_walk(name, &nd);
#else
	    err = path_lookup(name, LOOKUP_FOLLOW, &nd);
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	    if (err) {
		printk(KERN_ALERT "tracefs: error accessing hidden directory '%s'\n", name);
		hidden_root = ERR_PTR(err);
		goto out;
	    }
	    hidden_root = nd.dentry;
	    stopd(sb)->hidden_mnt = nd.mnt;
	    fist_dprint(6, "parse_options: new s_root, inode: 0x%lx, 0x%lx\n",
			(long) hidden_root, (long) hidden_root->d_inode);
	} else if (!strncmp("debug=", options, 6)) {
	    int debug = simple_strtoul(options + 6, NULL, 0);
	    fist_set_debug_value(debug);
	} else if (!strncmp("mmap=", options, 5)) {
		char mmap_tracing = simple_strtoul(options + 5, NULL, 0);
		if (mmap_tracing == 0) {
			mmap_support(sb) = 0;
		} else {
			mmap_support(sb) = 1;
		}
	} else {
	    printk(KERN_WARNING "tracefs: unrecognized option '%s'\n", options);
	}
	options = tmp + 1;
    }

out:
    print_exit_location();
    return hidden_root;
}

#ifdef FIST_MALLOC_DEBUG
/* for malloc debugging */
static atomic_t tracefs_malloc_counter;
static atomic_t tracefs_malloc_counter_diff;

void *
tracefs_kmalloc(size_t len, int flag, int line, const char *file)
{
    void *ptr = (void *) kmalloc(len, flag);
    if (ptr) {
	atomic_inc(&tracefs_malloc_counter);
	atomic_inc(&tracefs_malloc_counter_diff);
	printk(KERN_ALERT "KM:%d:%p:%d:%s\n", atomic_read(&tracefs_malloc_counter),ptr, line, file);
	//printk(KERN_ALERT "KM:%d:%d:%p:%d:%s\n", atomic_read(&tracefs_malloc_counter),atomic_read(&tracefs_malloc_counter_diff),ptr, line, file);
    }
    return ptr;
}

void
tracefs_kfree(void *ptr, int line, const char *file)
{
    atomic_inc(&tracefs_malloc_counter);
    atomic_dec(&tracefs_malloc_counter_diff);
    printk(KERN_ALERT "KF:%d:%p:%d:%s\n", atomic_read(&tracefs_malloc_counter),ptr, line, file);
    //printk(KERN_ALERT "KF:%d:%d:%p:%d:%s\n", atomic_read(&tracefs_malloc_counter),atomic_read(&tracefs_malloc_counter_diff),ptr, line, file);
    kfree(ptr);
}
#endif /* FIST_MALLOC_DEBUG */


/* for attach mode, we use a different ->read_super() in attach.c */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
static int
#else
super_block_t *
#endif
tracefs_read_super(super_block_t *sb, void *raw_data, int silent)
{
    super_block_t *ret_sb = NULL;
    dentry_t *hidden_root;
    int err = 0;

    print_entry_location();

#ifdef FIST_MALLOC_DEBUG
    atomic_set(&tracefs_malloc_counter, 0); /* for malloc debugging */
    atomic_set(&tracefs_malloc_counter_diff, 0); /* for malloc debugging */
#endif /* FIST_MALLOC_DEBUG */

    if (!raw_data) {
	printk(KERN_WARNING "tracefs_read_super: missing data argument\n");
	err = -EINVAL;
	goto out;
    }
    /*
     * Allocate superblock private data
     */
    stopdlh(sb) = KMALLOC(sizeof(struct tracefs_sb_info), GFP_KERNEL);
    if (!stopd(sb)) {
	printk(KERN_WARNING "%s: out of memory\n", __FUNCTION__);
	err = -ENOMEM;
	goto out;
    }
    memset(stopd(sb), 0, sizeof(struct tracefs_sb_info));
    stohs(sb) = NULL;
    stopnc(sb) = namecache_init();
    if (!stopnc(sb)) {
	printk(KERN_WARNING "%s: out of memory\n", __FUNCTION__);
	err = -ENOMEM;
	goto out_free;
    }

    mmap_support(sb) = 1; /* enable mmap support by default */


    hidden_root = tracefs_parse_options(sb, raw_data);
    if (IS_ERR(hidden_root)) {
	printk(KERN_WARNING "tracefs_read_super: lookup_dentry failed (err = %ld)\n", PTR_ERR(hidden_root));
	err = PTR_ERR(hidden_root);
	goto out_free;
    }
    if (!hidden_root->d_inode) {
	printk(KERN_WARNING "tracefs_read_super: no directory to interpose on\n");
	goto out_dput;
    }
    stohs(sb) = hidden_root->d_sb;

    /*
     * Linux 2.4.2-ac3 and beyond has code in
     * mm/filemap.c:generic_file_write() that requires sb->s_maxbytes
     * to be populated.  If not set, all write()s under that sb will
     * return 0.
     *
     * Linux 2.4.4+ automatically sets s_maxbytes to MAX_NON_LFS;
     * the filesystem should override it only if it supports LFS.
     */
    /* non-SCA code is good to go with LFS */
    sb->s_maxbytes = hidden_root->d_sb->s_maxbytes;

    sb->s_op = &tracefs_sops;
    /*
     * we can't use d_alloc_root if we want to use
     * our own interpose function unchanged,
     * so we simply replicate *most* of the code in d_alloc_root here
     */
    sb->s_root = d_alloc(NULL, &(const struct qstr) { name:"/", len:1, hash:0 });
    if (IS_ERR(sb->s_root)) {
	printk(KERN_WARNING "tracefs_read_super: d_alloc failed\n");
	err = -ENOMEM;
	goto out_dput;
    }

    sb->s_root->d_op = &tracefs_dops;
    sb->s_root->d_sb = sb;
    sb->s_root->d_parent = sb->s_root;
    /* link the upper and lower dentries */
    dtopdlh(sb->s_root) = (struct tracefs_dentry_info *) KMALLOC(sizeof(struct tracefs_dentry_info), GFP_KERNEL);
    if (!dtopd(sb->s_root)) {
	err = -ENOMEM;
	goto out_dput2;
    }
    dtohd(sb->s_root) = hidden_root;

    if (tracefs_interpose(hidden_root, sb->s_root, sb, 0))
	goto out_dput2;

    ret_sb = sb;
    fist_print_dentry("tracefs_read_super OUT hidden_dentry", hidden_root);
    fist_print_inode("tracefs_read_super OUT hidden_inode", hidden_root->d_inode);
    // next line causes null ptr deref at mount(2) time
    // fist_print_dentry("%s OUT sb->s_root", __FUNCTION__, sb->s_root);
    goto out;

 out_dput2:
    dput(sb->s_root);
 out_dput:
    dput(hidden_root);
 out_free:
    if (stopnc(sb)) {
	KFREE(stopnc(sb));
    }
    if (stopd(sb)) {
	mntput(stopd(sb)->hidden_mnt);
    }
    KFREE(stopd(sb));
    stopdlh(sb) = NULL;
 out:
    fist_print_sb("OUT sb", sb);
    if (stopd(sb))
	fist_print_sb("OUT hidden_sb", stohs(sb));
    print_exit_location();
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
    if(err)
	return ERR_PTR(err);
    return ret_sb;
#else
    return err;
#endif
}

/*----*/
// this structure *must* be static!!! (or the memory for 'name' gets
// corrupted in 2.3...)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
//No longer used in 2.6
static DECLARE_FSTYPE(tracefs_fs_type, "tracefs", tracefs_read_super, 0);
#else
static struct super_block *tracefs_get_sb(struct file_system_type *fs_type,
                                         int flags, const char *dev_name,
                                         void *raw_data) {
	return get_sb_nodev( fs_type, flags, raw_data, tracefs_read_super);
}

void tracefs_kill_block_super(struct super_block *sb)
{
    generic_shutdown_super(sb);
}

struct file_system_type tracefs_fs_type = {
    .name = "tracefs",
    .owner = THIS_MODULE,
    .fs_flags = 0,
    .get_sb = tracefs_get_sb,
    .kill_sb = tracefs_kill_block_super,
};
#endif //#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)

static int __init init_tracefs_fs(void)
{
    printk("Registering tracefs version $Id: main.c,v 1.26 2007/01/30 06:30:53 gopalas Exp $\n");
    return register_filesystem(&tracefs_fs_type);
}
static void __exit exit_tracefs_fs(void)
{
    printk("Unregistering tracefs version $Id: main.c,v 1.26 2007/01/30 06:30:53 gopalas Exp $\n");
    unregister_filesystem(&tracefs_fs_type);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
EXPORT_NO_SYMBOLS;
#endif
MODULE_AUTHOR("Erez Zadok <ezk@cs.columbia.edu>");
MODULE_DESCRIPTION("FiST-generated tracefs filesystem");

#ifdef TRACEFS_PROPRIETARY
MODULE_LICENSE("Proprietary");
#else
MODULE_LICENSE("GPL");
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,17)
MODULE_PARM(fist_debug_var, "i");
#else
module_param(fist_debug_var, int,0644);
#endif
MODULE_PARM_DESC(fist_debug_var, "Debug level");

module_init(init_tracefs_fs)
module_exit(exit_tracefs_fs)

/*
 * Local variables:
 * c-basic-offset: 4
 * End:
 */
