/*
 * Copyright (c) 2003-2007 Erez Zadok
 * Copyright (c) 2003-2007 Stony Brook University
 * Copyright (c) 2003-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2003-2005 Akshat Aranya
 * Copyright (c) 2006-2007 Gopala Surya
 */
/*
 * tracegen.c
 * implements a simple wrapper around the trace parsing library
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#define NEED_DESCRIPTIVE_MESSAGES
#include "messages.h"
#include "tracereader.h"
#include "namehash.h"
#include "pidhash.h"
#include "od_cksum.h"
#include <unistd.h>
#include "util.h"
static int indent_level = 0;

#define PID_INVALID (-1)
struct context {
    pid_t pid;
    char *pname;
    int len;
    int ret; /* return value for ops that return an integer */
    char *pargs;
    void *other;
    int other_fields_filled;
};

#define NAMECOMP_NUMFIELDS 3
struct namecomp {
   int parent_inum;
   int child_inum;
   char *name;
};

#define SYSCALL_NUMFIELDS 2
#define SYSCALL_NAME_LENGTH 64
struct syscall {
   unsigned short id;
   char name[SYSCALL_NAME_LENGTH];
};
#define MAX_SYSCALLS 256
static struct syscall *syscall_table[MAX_SYSCALLS];
static int nsyscalls;


static struct namehash *hash;
static struct pidhash *pid_hash;
#define PRINT_INDENT(...) ({int __j = indent_level; while (__j-- > 0) printf("  "); printf(__VA_ARGS__);})
void usage()
{
	fprintf(stderr, "Usage: tracegen [-c <key>] [-b <block-size>] [-d <key>] -z -k <key file> -f <file-name>\n");
	fprintf(stderr, "       -c <key> specifies that the file is checksummed\n");
      	fprintf(stderr, "       -b <block-size> specifies the block size of the checksum, default 4096\n");
	fprintf(stderr, "       -z decompress using zlib\n");
	fprintf(stderr, "       -d <key> specifies the decryption key\n");
	fprintf(stderr, "       -k <key file> specifies the key file for decrypting the trace fields\n");
	fprintf(stderr, "      -f specifies the file name\n");
	fprintf(stderr, "Note that order of -c and -z are important\n");
	exit(1);
}
void read_keys(char *keyfile)
{
	FILE *fp;
	char buf[256];
	struct decrypt_params keyp;
	int i;
	if ((fp = fopen(keyfile, "r")) == NULL) {
		fprintf(stderr, "Warning : could not open key file [%s] for reading\n", keyfile);
		return;
	}
	while (fgets(buf, 256, fp)) {
		char *name;
		char *key;
		buf[strlen(buf) - 1] = '\0';
		name = strtok(buf, ": ");
		key = strtok(NULL, ": ");
		if (name && key) {
			for (i = 0; i < ANON_MAX_FIELDS; i++) {
				if (strcmp(anon_keys[i].name, name) == 0) {
					if (make_key(&keyp, key)) {
						fprintf(stderr, "Warning: Invalid key\n");
						break;
					}
					anon_keys[i].valid = 1;
					memcpy(anon_keys[i].key, keyp.key, keyp.keylen);
					break;
				}
			}
			if (i == ANON_MAX_FIELDS) {
				fprintf(stderr, "Warning: Unknown field [%s]\n", name);
			}
		}
		else {
			fprintf(stderr, "Warning: Key file corrupt\n");
			fclose(fp);
			return;
		}
	}

	fclose(fp);
	return;


}
int main(int argc, char *argv[])
{
	int err;
	struct parse_params params;
	int c;
	int index;
	memset(&params, 0, sizeof(params));
	params.cparams.blk_size = CKSUM_BLOCK_SIZE;

	while ((c = getopt(argc, argv, "zc:b:f:d:k:")) != -1) {
		switch (c) {
			case 'c':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_CKSUM_MD5;
				if (hmac_make_key(&params.cparams, optarg)) {
					fprintf(stderr, "Invalid key specified\n");
					usage();
				}
				break;
			case 'b':
				params.cparams.blk_size = atoi(optarg);
				break;
			case 'z':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_UNCOMPRESS;
				break;
			case 'f':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_FILE;
				params.file_name = optarg;
				break;
			case 'd':
				if (params.n_ops == MAX_OPS) {
					fprintf(stderr, "Too many operations specified\n");
					usage();
				}
				params.ops[params.n_ops++] = OP_DECRYPT;
				if (make_key(&params.dparams, optarg)) {
					fprintf(stderr, "Invalid key specified\n");
					usage();
				}
				break;
			case 'k':
				read_keys(optarg);
				break;
			case '?':
				fprintf(stderr, "Unknown option `-%c'.\n\n", optopt);
				usage();
				break;
			default:
				usage();
				break;
		}
	}
	index = optind;
	if (!params.file_name) {
		fprintf(stderr, "File name not specified\n");
		usage();
	}
	if ((hash = create_namehash(1024)) == NULL) {
		fprintf(stderr, "Could not create name hash\n");
		exit(2);
	}

	if ((pid_hash = pidhash_create(1024)) == NULL) {
		fprintf(stderr, "Could not create PID hash\n");
		exit(3);
	}
	memset(syscall_table, 0, sizeof(syscall_table));

	err = parse(&params);

	if (err != ERR_OK)
		fprintf(stderr, "Parsing failed with error code %d\n", err);

	pidhash_destroy(pid_hash);

	destroy_namehash(hash);
	return err;
}

struct descriptor {
	op_t type;
	char *str_type;
};

void *on_start_trace_message(op_t type, int pre_or_post,void* context)
{
	struct context *cont = malloc(sizeof(struct context));
	PRINT_INDENT("BEGIN: %s %s\n", type_to_str_type(type), pre_or_post ? "POST_OP " : "PRE_OP");
	indent_level++;
	memset(cont, 0, sizeof(*cont));
	cont->pid = PID_INVALID;
	cont->ret = -1;

	if (pre_or_post && (type == OP_LOOKUP || type == OP_LINK || type == OP_MKDIR)) {
		struct namecomp *tmp = malloc(sizeof(struct namecomp));
		memset(tmp, 0, sizeof(*tmp));
		cont->other = (void *)tmp;
	} else if (type == MSG_SYSCALL) {
		struct syscall *tmp = malloc(sizeof(struct syscall));
		memset(tmp, 0, sizeof(*tmp));
		cont->other = (void *)tmp;
	}
	return cont;
}

int on_end_trace_message(op_t type, int pre_or_post, void *cont)
{
    struct context *context = cont;
    if (context) {
	if (pre_or_post && (type == OP_LOOKUP || type == OP_LINK || type == OP_MKDIR)) {
	    struct namecomp *tmp = (struct namecomp *)context->other;
	    if (context->ret == 0 && context->other_fields_filled == NAMECOMP_NUMFIELDS) {
		add_entry(hash, tmp->child_inum, tmp->parent_inum, tmp->name);
	    }
	    if (tmp->name)
		free(tmp->name);
	    free(tmp);
	} else if (type == MSG_SYSCALL && context->other_fields_filled == SYSCALL_NUMFIELDS) {
		syscall_table[nsyscalls++] = (struct syscall *)context->other;
	}
    }
    if (context->pid != PID_INVALID) {
	struct pidhash_entry *entry = pidhash_lookup(pid_hash, context->pid);
	if (context->pname)
	    entry = pidhash_add_entry(pid_hash, context->pid, context->pname, context->len, entry);
	if (entry) {
	    int i;
	    PRINT_INDENT("Process Name: \"");
	    for (i = 0; i < entry->len; i++) printf("%c", entry->pname[i]);
	    printf("\"\n");
	} else
	    PRINT_INDENT("PID = %d\n", context->pid);
    }
    indent_level--;
    PRINT_INDENT("END: %s %s\n", type_to_str_type(type), pre_or_post ? "POST_OP" : "PRE_OP");

    if (context->pname)
	free(context->pname);
    free(context);
    return 0;
}

void on_error(int error_type, int msg_num, char *desc)
{
	fprintf(stderr, "Error occurred while parsing trace: err_num = %d, msg_num = %d, description = %s\n", error_type, msg_num, desc);
}

void on_warning(int warn_type, char *desc)
{
    fprintf(stderr, "Warning: %s [%d]\n", desc, warn_type);
}

extern int get_decrypted_value(int anon_type, void *invalue, int inlen, void **outvalue, int *outl);
extern void release_decrypted_value(void *dvalue);


int on_data_field(op_t op, int pre_or_post, int tlv, field_t type, int length, void *value, void *cont)
{
	struct context *context = cont;
	void *dvalue = NULL;
	int err = ERR_OK;
	int anon_type = ANON_MAX_FIELDS;
	switch (type) {
		case ARG_UID:
			if (anon_keys[ANON_UID].valid) anon_type = ANON_UID;
			break;
		case ARG_GID:
			if (anon_keys[ANON_GID].valid) anon_type = ANON_GID;
			break;
		case ARG_PID:
			if (anon_keys[ANON_PID].valid) anon_type = ANON_PID;
			break;
		case ARG_PNAME:
			if (anon_keys[ANON_PNAME].valid) anon_type = ANON_PNAME;
			break;
		case ARG_PATHNAME:
			if (anon_keys[ANON_PATHNAME].valid) anon_type = ANON_PATHNAME;
			break;
	}
	if (anon_type != ANON_MAX_FIELDS) {
		err = get_decrypted_value(anon_type, value, length, &dvalue, &length);
		if (err != ERR_OK)
			goto out;
		value = dvalue;
	}

	switch (type) {
		case ARG_TIMESTAMP:
			{
				unsigned long secs;
				memcpy(&secs, value, sizeof(secs));
				PRINT_INDENT("Time: %s\n", ctime((const time_t*)&secs));
			}
			break;
		case ARG_UTIMESTAMP:
			{
				unsigned long secs, usecs;
				memcpy(&secs, value, sizeof(secs));
				memcpy(&usecs, value + sizeof(secs), sizeof(usecs));
				PRINT_INDENT("UTime: %ld, %ld\n", secs, usecs);
			}
			break;
		case ARG_INUM:
			{
				unsigned int inum;
				struct namehash_entry *entry;
				memcpy(&inum, value, sizeof(inum));
				PRINT_INDENT("INum: %d ", inum);
				entry = lookup(hash, inum);
				if (entry) {
					struct names *tmp = entry->namelist;
					printf("Names: ");
					while (tmp) {
						printf("%s ", tmp->name);
						tmp = tmp->next;
					}
				}
				printf("\n");
				if (pre_or_post && context && context->other) {
					if (op == OP_LOOKUP || op == OP_MKDIR) {
						struct namecomp *tmp = (struct namecomp *)context->other;
						tmp->parent_inum = inum;
						context->other_fields_filled++;
					} else if (op == OP_LINK) {
						struct namecomp *tmp = (struct namecomp *)context->other;
						tmp->child_inum = inum;
						context->other_fields_filled++;
					}
				}

			}
			break;
		case ARG_NEW_INUM:
			{
				unsigned int inum;
				memcpy(&inum, value, sizeof(inum));
				PRINT_INDENT("INum: %d\n", inum);
			}
			break;
		case ARG_CHILD_INUM:
			{
				unsigned int inum;
				memcpy(&inum, value, sizeof(inum));
				PRINT_INDENT("Child INum: %d\n", inum);
				if ((op == OP_LOOKUP || op == OP_MKDIR) && pre_or_post && context && context->other) {
					struct namecomp *tmp = (struct namecomp *)context->other;
					tmp->child_inum = inum;
					context->other_fields_filled++;
				}
			}
			break;

		case ARG_PATHNAME:
			{
				char *name = malloc(length + 1);
				memcpy(name, value, length);
				name[length] = 0;
				PRINT_INDENT("File name: %s\n", name);
				if ((op == OP_LOOKUP || op == OP_MKDIR) && pre_or_post && context) {
					struct namecomp *tmp = (struct namecomp *)context->other;
					tmp->name = malloc(length+1);
					strcpy(tmp->name, name);
					context->other_fields_filled++;
				}
				free(name);
			}
			break;
		case ARG_STR:
			{
				char *name = malloc(length + 1);
				memcpy(name, value, length);
				name[length] = 0;
				PRINT_INDENT("String: %s\n", name);
				free(name);
			}
			break;
		case ARG_ULONG:
			{
				unsigned long tmp;
				memcpy(&tmp, value, sizeof(tmp));
				printf("flags = %lu\n", tmp);
			}
			break;
		case ARG_NEW_PATHNAME:
			{
				char *name = malloc(length + 1);
				memcpy(name, value, length);
				name[length] = 0;
				PRINT_INDENT("New file name: %s\n", name);
				if (op == OP_LINK && pre_or_post && context && context->other) {
					struct namecomp *tmp = (struct namecomp *)context->other;
					tmp->name = malloc(length+1);
					strcpy(tmp->name, name);
					context->other_fields_filled++;
				}
				free(name);
			}
			break;

			/*
			   case ARG_SB_MAJOR:
			   {
			   unsigned int dev;
			   memcpy(&dev, value, sizeof(dev));
			   PRINT_INDENT("Major dev #: %ud", dev);
			   }
			   break;
			   case ARG_SB_MINOR:
			   {
			   unsigned int dev;
			   memcpy(&dev, value, sizeof(dev));
			   PRINT_INDENT("Minor dev #: %ud", dev);
			   }
			   break;
			 */
		case ARG_INT_FLAGS:
			{
				unsigned int flags;
				memcpy(&flags, value, sizeof(flags));
				PRINT_INDENT("Flags: %ud\n", flags);
			}
			break;
		case ARG_UID:
			{
				uid_t uid;
				memcpy(&uid, value, sizeof(uid));
				PRINT_INDENT("UID: %d\n", uid);
			}
			break;
		case ARG_GID:
			{
				gid_t gid;
				memcpy(&gid, value, sizeof(gid));
				PRINT_INDENT("GID: %d\n", gid);
			}
			break;
		case ARG_PID:
			{
				pid_t pid;
				memcpy(&pid, value, sizeof(pid));
				context->pid = pid;
				//PRINT_INDENT("PID: %d\n", pid);
			}
			break;
		case ARG_DENTRY_PTR:
			{
				unsigned int ptr;
				memcpy(&ptr, value, sizeof(ptr));
				PRINT_INDENT("Dentry PTR: %p\n", (void *)ptr);
			}
			break;
		case ARG_INODE_PTR:
			{
				unsigned int ptr;
				memcpy(&ptr, value, sizeof(ptr));
				PRINT_INDENT("Inode PTR: %p\n", (void *)ptr);
			}
			break;
		case ARG_FILE_PTR:
			{
				unsigned int ptr;
				memcpy(&ptr, value, sizeof(ptr));
				PRINT_INDENT("File PTR: %p\n", (void *)ptr);
			}
			break;
		case ARG_NEW_INODE_PTR:
			{
				unsigned int ptr;
				memcpy(&ptr, value, sizeof(ptr));
				PRINT_INDENT("Inode2 PTR: %p\n", (void *)ptr);
			}
			break;
		case ARG_NEW_DENTRY_PTR:
			{
				unsigned int ptr;
				memcpy(&ptr, value, sizeof(ptr));
				PRINT_INDENT("Dentry2 PTR: %p\n", (void *)ptr);
			}
			break;
		case ARG_FILE_FLAGS:
			{
				unsigned int flags;
				memcpy(&flags, value, sizeof(flags));
				PRINT_INDENT("File flags 0x%x\n", flags);
			}
			break;
		case ARG_LOFFSET:
			{
				loff_t offset;
				memcpy(&offset, value, sizeof(offset));
				PRINT_INDENT("Offset: %lld\n", offset);
			}
			break;
		case ARG_REQUESTED_SIZE:
			{
				size_t size;
				memcpy(&size, value, sizeof(size));
				PRINT_INDENT("Requested size = %d\n", size);
			}
			break;
		case ARG_RET:
			{
				int ret;
				memcpy(&ret, value, sizeof(ret));
				context->ret = ret;
				PRINT_INDENT("Return = %d\n", ret);
			}
			break;
		case ARG_MSG_ID:
			{
				op_t ret;
				memcpy(&ret, value, sizeof(ret));
				PRINT_INDENT("MSG ID = %s ", type_to_str_type(ret));
			}
			break;
		case ARG_MSG_DESC:
			{
				char *name = malloc(length + 1);
				memcpy(name, value, length);
				name[length] = 0;
				PRINT_INDENT("MSG Name: %s\n", name);
				free(name);
			}
			break;
		case ARG_FIELD_ID:
			{
				field_t ret;
				memcpy(&ret, value, sizeof(ret));
				PRINT_INDENT("FIELD ID = %d ", ret);
			}
			break;
		case ARG_FIELD_LEN:
			{
				field_len_t len;
				memcpy(&len, value, sizeof(len));
				PRINT_INDENT("FIELD len = %d ", len);
			}
			break;
		case ARG_FIELD_DESC:
			{
				char *name = malloc(length + 1);
				memcpy(name, value, length);
				name[length] = 0;
				PRINT_INDENT("Field Name: %s\n", name);
				free(name);
			}
			break;

		case ARG_SYSCALL:
			{
				unsigned short id;
				memcpy(&id, value, sizeof(id));
				if (op != MSG_SYSCALL) {
					int i;
					for (i = 0; i < nsyscalls; i++) {
						if (syscall_table[i]->id == id)
							break;
					}
					PRINT_INDENT("Syscall name = %s ", (i < nsyscalls) ? syscall_table[i]->name : "UNKNOWN");
					printf("[%d]\n", id);
				} else {
					if (context && context->other) {
						((struct syscall *)context->other)->id = id;
						context->other_fields_filled++;
					}
				}
			}
			break;

		case ARG_SYSCALL_NAME:
			{
				if (op == MSG_SYSCALL && context && context->other) {
					length = (length > SYSCALL_NAME_LENGTH - 1) ? SYSCALL_NAME_LENGTH - 1 : length;
					memcpy(((struct syscall *)context->other)->name, value, length);
					context->other_fields_filled++;
				}
			}

		case ARG_PNAME:
			{
				char *name = malloc(length);
				memcpy(name, value, length);
				//name[length] = 0;
				context->pname = name;
				context->len = length;
				//PRINT_INDENT("Process Name: %s\n", name);
				//free(name);
			}
			break;
		case ARG_SYSNAME:
		case ARG_RELEASE:
		case ARG_VERSION:
		case ARG_MACHINE:
			{
				PRINT_INDENT("%s: ", get_str_arg_type(type));
				printf("%s\n", (char *)value);
			}
			break;
		case ARG_DIR_INUM:
			{
				unsigned int inum;
				memcpy(&inum, value, sizeof(inum));
				PRINT_INDENT("Dir INum: %d\n", inum);
				if (op == OP_LINK && pre_or_post && context && context->other) {
					struct namecomp *tmp = (struct namecomp *)context->other;
					tmp->parent_inum = inum;
					context->other_fields_filled++;
				}
			}
			break;

		default:
			{
				int i;
				PRINT_INDENT("Arg type: %s ", get_str_arg_type(type));
				for (i = 0; i < length; i++)
					PRINT_INDENT(" 0x%x", (unsigned int)(((char *)value)[i]));
				printf("\n");
			}
	}
out:
	if (dvalue)
		release_decrypted_value(dvalue);
	return err;
}

int on_header(struct stream_header *header)
{
	// do nothing
	return ERR_OK;
}

