/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2004-2005 Timothy Wong
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include "debug.h"
#include "hash_list.h"

struct list_entry_s {
  struct list_entry_s* next;
  struct list_entry_s* prev;
  void* key;
  int keylen;
  void* value;
};


static int _add_to_list(hash_list_t* hash_list,list_entry_t* next,list_entry_t* prev,void* key,int keylen,void* value)
{
  int err = 0;
  list_entry_t* new_entry;
  print_entry_location();
  assert(hash_list != NULL);
  assert(keylen > 0);

  new_entry = malloc(sizeof(list_entry_t));
  if(!new_entry) {
    print_error_nomem();
    exit(ENOMEM);
  }

  memset(new_entry,0,sizeof(list_entry_t));
  new_entry->next = next;
  new_entry->prev = prev;
  new_entry->keylen = keylen;
  new_entry->value = value;

  new_entry->key = malloc(keylen);
  if(!new_entry->key) {
    print_error_nomem();
    exit(ENOMEM);
  }
  memcpy(new_entry->key,key,keylen);

  err = hash_add_allow_dupes(&hash_list->hash,key,keylen,new_entry);
  if(err)
    goto out;
  if(next)
    new_entry->next->prev = new_entry;
  else if(prev)
    hash_list->list_tail = new_entry;
  if(prev)
    new_entry->prev->next = new_entry;
  else if(next)
    hash_list->list_head = new_entry;

  if(!hash_list->list_head && !hash_list->list_tail) {
    hash_list->list_head = new_entry;
    hash_list->list_tail = new_entry;
  }

  hash_list->list_size++;

 out:
  print_exit_location();
  return err;
}

static void* _remove_from_list(hash_list_t* hash_list,list_entry_t* list_entry)
{
  void* ret = NULL;
  print_entry_location();
  assert(hash_list != NULL);
  assert(list_entry != NULL);

  hash_remove(&hash_list->hash,list_entry->key,list_entry->keylen);
  if(list_entry->prev != NULL)
    list_entry->prev->next = list_entry->next;
  else if(list_entry->next != NULL)
    hash_list->list_head = list_entry->next;
  if(list_entry->next != NULL)
    list_entry->next->prev = list_entry->prev;
  else if(list_entry->prev != NULL)
    hash_list->list_tail = list_entry->prev;
  else {
    //both next and prev are NULL
    hash_list->list_head = NULL;
    hash_list->list_tail = NULL;
  }

  ret = list_entry->value;
  free(list_entry->key);
  free(list_entry);

  hash_list->list_size--;

  print_exit_location();
  return ret;
}

int hash_list_init(hash_list_t* hash_list,int hash_size)
{
  int err = 0;
  print_entry_location();
  assert(hash_list != NULL);

  memset(hash_list,0,sizeof(list_entry_t));
  hash_init(&hash_list->hash,hash_size);

  print_exit_location();
  return err;
}

inline int hash_list_get_size(hash_list_t* hash_list)
{
  assert(hash_list != NULL);
  return hash_list->list_size;
}

inline int hash_list_add(hash_list_t* hash_list,void* key,int keylen,void* value)
{
  return hash_list_add_first(hash_list,key,keylen,value);
}

//Adds to the end of the list
int hash_list_add_first(hash_list_t* hash_list,void* key,int keylen,void* value)
{
  int err = 0;
  print_entry_location();
  assert(hash_list != NULL);

  _add_to_list(hash_list,hash_list->list_head,NULL,key,keylen,value);

  print_exit_location();
  return err;
}

int hash_list_add_last(hash_list_t* hash_list,void* key,int keylen,void* value)
{
  int err = 0;
  print_entry_location();
  assert(hash_list != NULL);

  _add_to_list(hash_list,NULL,hash_list->list_tail,key,keylen,value);

  print_exit_location();
  return err;
}

//returns Null on fail
void* hash_list_remove(hash_list_t* hash_list,void* key,int keylen)
{
  void* ret = NULL;
  list_entry_t* list_entry;
  print_entry_location();
  assert(hash_list != NULL);

  list_entry = hash_find(&hash_list->hash,key,keylen);
  if(!list_entry)
    goto out;
  ret = _remove_from_list(hash_list,list_entry);

 out:
  print_exit_location();
  return ret;
}

//Returns and removes the first element
void* hash_list_remove_first(hash_list_t* hash_list)
{
  void* ret = NULL;
  print_entry_location();
  assert(hash_list != NULL);

  if(hash_list->list_head)
    ret = _remove_from_list(hash_list,hash_list->list_head);

  print_exit_location();
  return ret;
}

//Returns and removes the last element
void* hash_list_remove_last(hash_list_t* hash_list)
{
  void* ret = NULL;
  print_entry_location();
  assert(hash_list != NULL);

  if(hash_list->list_tail)
    ret = _remove_from_list(hash_list,hash_list->list_tail);

  print_exit_location();
  return ret;
}

void* hash_list_remove_location(hash_list_t* hash_list,int loc)
{
  void* ret = NULL;
  list_entry_t* ptr;
  print_entry_location();
  assert(hash_list != NULL);

  ptr = hash_list->list_head;
  while(ptr != NULL && loc-- > 0)
    ptr = ptr->next;

  if(ptr)
    ret = _remove_from_list(hash_list,ptr);

  print_exit_location();
  return ret;
}

//Returns the first element
void* hash_list_get_first(hash_list_t* hash_list)
{
  void* ret = NULL;
  print_entry_location();
  assert(hash_list != NULL);

  if(hash_list->list_head != NULL)
    ret = hash_list->list_head->value;

  print_exit_location();
  return ret;
}

//Returns the last element
void* hash_list_get_last(hash_list_t* hash_list)
{
  void* ret = NULL;
  print_entry_location();
  assert(hash_list != NULL);

  if(hash_list->list_tail)
    ret = hash_list->list_tail->value;

  print_exit_location();
  return ret;
}

void* hash_list_get_location(hash_list_t* hash_list,int loc)
{
  void* ret = NULL;
  list_entry_t* ptr;
  print_entry_location();
  assert(hash_list != NULL);

  ptr = hash_list->list_head;
  while(ptr != NULL && loc-- > 0)
    ptr = ptr->next;

  if(ptr)
    ret = ptr->value;

  print_exit_location();
  return ret;
}

void* hash_list_find(hash_list_t* hash_list,void* key,int keylen)
{
  void* ret = NULL;
  list_entry_t* ptr;
  print_entry_location();
  assert(hash_list != NULL);

  ptr = hash_find(&hash_list->hash,key,keylen);

  if(ptr)
    ret = ptr->value;

  print_exit_location();
  return ret;
}

int hash_list_destroy(hash_list_t* hash_list)
{
  int err = 0;
  list_entry_t* ptr;
  list_entry_t* prev;
  print_entry_location();
  assert(hash_list != NULL);

  hash_destroy(&hash_list->hash);
  ptr = hash_list->list_head;

  while(ptr != NULL){
    prev = ptr;
    ptr = ptr->next;
    free(prev->key);
    free(prev);
  }

  memset(hash_list,0,sizeof(hash_list_t));

  print_exit_location();
  return err;
}
