/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2004-2005 Timothy Wong
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <assert.h>
#include "debug.h"
#include "hashed_heap.h"

struct heap_entry_s {
  struct heap_entry_s* parent;
  struct heap_entry_s* left_child;
  struct heap_entry_s* right_child;
  int descendents; //Number of nodes beneath
  int keylen;
  void* key;
  void* value;
};

#define TRANSPOSE_HEAP_POINTER(ENTRY1,ENTRY1_PTR,ENTRY2,ENTRY2_PTR) do { \
    heap_entry_t* ptr; \
    if(ENTRY1_PTR == ENTRY2) { \
       ptr = ENTRY2_PTR; \
       ENTRY2_PTR = ENTRY1; \
       ENTRY1_PTR = ptr; \
    } \
    else if(ENTRY2_PTR == ENTRY1) { \
       ptr = ENTRY1_PTR; \
       ENTRY1_PTR = ENTRY2; \
       ENTRY2_PTR = ptr; \
    } \
    else { \
       ptr = ENTRY1_PTR; \
       ENTRY1_PTR = ENTRY2_PTR; \
       ENTRY2_PTR = ptr; \
    }  \
}while(0)

/*
 * Swap Pointers (used to flip parent and child)
 * Caller is responsible for fixing root and last pointers in hashed_heap obj.
 */
static int _transpose_nodes(hashed_heap_t* hashed_heap,heap_entry_t *entry1,heap_entry_t *entry2)
{
  int err = 0;
  heap_entry_t *ptr;
  int descendents;
  print_entry_location();
  assert(entry1);
  assert(entry2);

  //Transpose root if one of the nodes is root
  if(entry1 == hashed_heap->heap_root)
    hashed_heap->heap_root = entry2;
  else if(entry2 == hashed_heap->heap_root)
    hashed_heap->heap_root = entry1;

  //Transpose Parent
  TRANSPOSE_HEAP_POINTER(entry1,entry1->parent,entry2,entry2->parent);
  /* Now we need to fix the rest of the tree */
  if(entry2->parent && entry2->parent != entry1) {
    if(entry2->parent->left_child == entry1)
      entry2->parent->left_child = entry2;
    else
      entry2->parent->right_child = entry2;
  }
  if(entry1->parent && entry1->parent != entry2) {
    if(entry1->parent->left_child == entry2)
      entry1->parent->left_child = entry1;
    else
      entry1->parent->right_child = entry1;
  }

  //Transpose Left Child
  TRANSPOSE_HEAP_POINTER(entry1,entry1->left_child,entry2,entry2->left_child);
  /* Now we need to fix the rest of the tree */
  if(entry2->left_child && entry2->left_child != entry1)
    entry2->left_child->parent = entry2;
  if(entry1->left_child && entry1->left_child != entry2)
    entry1->left_child->parent = entry1;

  //Transpose Right Child
  TRANSPOSE_HEAP_POINTER(entry1,entry1->right_child,entry2,entry2->right_child);
  /* Now we need to fix the rest of the tree */
  if(entry2->right_child && entry2->right_child != entry1)
    entry2->right_child->parent = entry2;
  if(entry1->right_child && entry1->right_child != entry2)
    entry1->right_child->parent = entry1;

  //Transpose the Number of Descendents
  descendents = entry1->descendents;
  entry1->descendents = entry2->descendents;
  entry2->descendents = descendents;

  print_exit_location();
  return err;
}

static int _decrement_descendents(hashed_heap_t* hashed_heap,heap_entry_t* entry)
{
  int err = 0;
  print_entry_location();

  if(entry) {
    err = _decrement_descendents(hashed_heap,entry->parent);
    entry->descendents--;
  }

  print_exit_location();
  return 0;
}

//After Insert - Not using recursion for speed
static int _heapify_up(hashed_heap_t *hashed_heap,heap_entry_t *entry)
{
  int err = 0;
  print_entry_location();
  assert(hashed_heap);
  assert(entry);

  while(entry->parent &&
	(hashed_heap->compare_fn(entry->key,entry->keylen,
				 entry->parent->key,entry->parent->keylen) > 0))
    _transpose_nodes(hashed_heap,entry,entry->parent);

  assert(entry != entry->parent && entry != entry->left_child && entry != entry->right_child);

 out:
  print_exit_location();
  return err;
}

//After remove - Not using recursion for speed
static int _heapify_down(hashed_heap_t* hashed_heap,heap_entry_t* entry)
{
  int err = 0;
  heap_entry_t *child_entry;
  print_entry_location();
  assert(hashed_heap);
  assert(entry);

  /*
   * This can happen if remove op finds longest tree node to replace removed
   * node; since it's from different branches, this can occur!
   */
  if(entry->parent && hashed_heap->compare_fn(entry->key,
					      entry->keylen,
					      entry->parent->key,
					      entry->parent->keylen)) {
    err = -1;
    goto out;
  }

  //If one branch is empty, then stop and fill that one!
  while(entry->left_child || entry->right_child) {
    //Find larger child
    if(!entry->left_child)
      child_entry = entry->right_child;
    else if(!entry->right_child)
      child_entry = entry->left_child;
    else {
      //Compare the two children to see which is larger
      if(hashed_heap->compare_fn(entry->left_child->key,
				 entry->left_child->keylen,
				 entry->right_child->key,
				 entry->right_child->keylen) > 0)
	child_entry = entry->left_child;
      else
	child_entry = entry->right_child;
    }

    //Compare entry with largest child
    if(hashed_heap->compare_fn(entry->key,entry->keylen,
			       child_entry->key,child_entry->keylen) >= 0)
      goto done; //Done, entry is larger

    //Swap
    _transpose_nodes(hashed_heap,entry,child_entry);
 }

done:
  assert(entry != entry->parent && entry != entry->left_child && entry != entry->right_child);

out:
  print_exit_location();
  return err;
}

/*
 * Find the leaf with the smallest branch - increases descendent count
 */
heap_entry_t* _insert_to_bottom(hashed_heap_t* hashed_heap,heap_entry_t* entry)
{
  heap_entry_t* parent_entry;
  print_entry_location();

  if(!hashed_heap->heap_root) {
    hashed_heap->heap_root = entry;
    goto out;
  }

  //Find the Leaf with the smallest height
  parent_entry = hashed_heap->heap_root;
  while(parent_entry->left_child && parent_entry->right_child) {
    parent_entry->descendents++;
    if(parent_entry->left_child->descendents <= parent_entry->right_child->descendents)
      parent_entry = parent_entry->left_child;
    else
      parent_entry = parent_entry->right_child;
  }
  parent_entry->descendents++;
  entry->parent = parent_entry;
  if(!parent_entry->left_child)
    parent_entry->left_child = entry;
  else if(!parent_entry->right_child)
    parent_entry->right_child = entry;

 out:
  print_exit_location();
  return entry;
}

/*
 * Find the leaf with the largest branch - decreases descendent count
 */
heap_entry_t* _remove_from_bottom(hashed_heap_t* hashed_heap)
{
  heap_entry_t* entry = NULL;
  print_entry_location();

  if(!hashed_heap->heap_root)
    goto out;

  //Find the leaf with the largest height
  entry = hashed_heap->heap_root;
  while(entry->left_child || entry->right_child) {
    entry->descendents--;
    if(!entry->left_child)
      entry = entry->right_child;
    else if(!entry->right_child)
      entry = entry->left_child;
    else {
      if(entry->left_child->descendents > entry->right_child->descendents)
	entry = entry->left_child;
      else
	entry = entry->right_child;
    }
  }

  //Remove from Tree
  if(entry->parent) {
    if(entry->parent->left_child == entry)
      entry->parent->left_child = NULL;
    else
      entry->parent->right_child = NULL;
    entry->descendents = 0;
    entry->parent = NULL;
  }

 out:
  print_exit_location();
  return entry;
}

int hashed_heap_init(hashed_heap_t* hashed_heap,int hash_size,int (*compare_fn)(void*,int,void*,int))
{
  int err = 0;
  print_entry_location();
  assert(hashed_heap);
  assert(compare_fn);

  memset(hashed_heap,0,sizeof(hashed_heap_t));
  hash_init(&hashed_heap->hash,hash_size);
  hashed_heap->compare_fn = compare_fn;

  print_exit_location();
  return err;
}

inline int hashed_heap_get_size(hashed_heap_t* hashed_heap)
{
  assert(hashed_heap);
  return hashed_heap->hash.number_of_objects;
}

int hashed_heap_add(hashed_heap_t* hashed_heap,void* key,int keylen,void* value)
{
  int err = 0;
  heap_entry_t* entry;
  print_entry_location();
  assert(hashed_heap);
  assert(key);
  assert(keylen > 0);

  entry = malloc(sizeof(heap_entry_t));
  if(!entry) {
    print_error_nomem();
    exit(ENOMEM);
  }
  memset(entry,0,sizeof(heap_entry_t));
  entry->key = malloc(keylen);

  //Copy Values into Entry
  memcpy(entry->key,key,keylen);
  entry->keylen = keylen;
  entry->value = value;
  entry->descendents = 0;

  assert(!hash_add(&hashed_heap->hash,key,keylen,entry));
  entry = _insert_to_bottom(hashed_heap,entry);

  if(entry != hashed_heap->heap_root)
    _heapify_up(hashed_heap,entry);

  print_exit_location();
  return err;
}

void* hashed_heap_remove(hashed_heap_t* hashed_heap,void* key,int keylen)
{
  void* value = NULL;
  heap_entry_t *entry;
  heap_entry_t *replacement_entry;
  print_entry_location();
  assert(hashed_heap);

  entry = hash_remove(&hashed_heap->hash,key,keylen);
  if(!entry)
    goto out;

  replacement_entry = _remove_from_bottom(hashed_heap);
  if(replacement_entry == hashed_heap->heap_root) {
    hashed_heap->heap_root = NULL;
    goto set_value_and_free;
  }
  if(replacement_entry == entry)
    goto set_value_and_free;

  //Transpose nodes
  _transpose_nodes(hashed_heap,replacement_entry,entry);
  //Fix child references
  if(replacement_entry->left_child)
    replacement_entry->left_child->parent = replacement_entry;
  if(replacement_entry->right_child)
    replacement_entry->right_child->parent = replacement_entry;
  //Re-Heap - if heapify_down() fails, means replacement is larger than parent
  if(_heapify_down(hashed_heap,replacement_entry))
    _heapify_up(hashed_heap,replacement_entry);

 set_value_and_free:
  value = entry->value;
  free(entry->key);
  free(entry);

 out:
  print_exit_location();
  return value;
}

void* hashed_heap_remove_root(hashed_heap_t* hashed_heap)
{
  void* value = NULL;
  print_entry_location();
  assert(hashed_heap);

  if(!hashed_heap->heap_root)
    goto out;

  value = hashed_heap_remove(hashed_heap,
			     hashed_heap->heap_root->key,
			     hashed_heap->heap_root->keylen);
  assert(value);

 out:
  print_exit_location();
  return value;
}

void* hashed_heap_get_root(hashed_heap_t* hashed_heap)
{
  void* value = NULL;
  print_entry_location();

  if(!hashed_heap->heap_root)
    goto out;

  value = hashed_heap_find(hashed_heap,
			   hashed_heap->heap_root->key,
			   hashed_heap->heap_root->keylen);
  assert(value);

 out:
  print_exit_location();
  return value;
}

void* hashed_heap_find(hashed_heap_t* hashed_heap,void* key,int keylen)
{
  void* value;
  print_entry_location();
  assert(hashed_heap);
  assert(key && keylen > 0);

  value = hash_find(&hashed_heap->hash,key,keylen);

  print_exit_location();
  return value;
}

int hashed_heap_destroy(hashed_heap_t* hashed_heap)
{
  int err = 0;
  heap_entry_t *left_child, *right_child;
  print_entry_location();
  assert(hashed_heap);

  if(!hashed_heap->heap_root)
    goto out;

  //Set child pointers
  left_child = hashed_heap->heap_root->left_child;
  right_child = hashed_heap->heap_root->right_child;

  //Recursively destroy left branch
  hashed_heap->heap_root = left_child;
  hashed_heap_destroy(hashed_heap);

  //Recursively destroy right branch
  hashed_heap->heap_root = right_child;
  hashed_heap_destroy(hashed_heap);

  //Free Heap Entry
  if(hashed_heap->heap_root && hashed_heap->heap_root->key)
    free(hashed_heap->heap_root->key);
  if(hashed_heap->heap_root)
    free(hashed_heap->heap_root);

  out:
  print_exit_location();
  return err;
}
