
/*
 * Copyright (c) 2004-2007 Erez Zadok
 * Copyright (c) 2004-2007 Stony Brook University
 * Copyright (c) 2004-2007 The Research Foundation of State University of New York
 * Copyright (c) 2004-2005 Timothy Wong
 * Copyright (c) 2004-2007 Nikolai Joukov
 * Copyright (c) 2004-2005 Charles Wright
 * Copyright (c) 2006-2007 Gopala Surya
 *
 * For specific licensing information, see the COPYING file distributed with
 * this package, or get one from ftp://ftp.filesystems.org/pub/fist/COPYING.
 *
 * This Copyright notice must be kept intact and distributed with all
 * fistgen sources INCLUDING sources generated by fistgen.
 */

#define NEED_DESCRIPTIVE_MESSAGES
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <linux/limits.h>
#include <assert.h>
#include <linux/errno.h>
#include "md5.h"
#include "replay.h"
#include "res_manager.h"
#include "tracereader.h"
#include "util.h"
#include "hash.h"
#include "hash_list.h"
#include "thread_manager.h"
#include "hashed_heap.h"
#include "debug.h"

#ifndef O_LARGEFILE
#define O_LARGEFILE 0100000
#endif

#define PAGE_SIZE 4096 //Page Sizes for RAT ops to write(2).
#define MD5_CKSUM_SIZE 16

#define _ADD_TO_LIST(HEAD,TAIL,TAIL_NEXT,ENTRY,SIZE) \
	ENTRY = malloc(SIZE); \
	if(ENTRY == NULL) { \
		print_error_nomem(); \
		exit(ENOMEM); \
	} \
	memset(ENTRY,0,SIZE); \
	if(HEAD) \
		TAIL_NEXT = ENTRY; \
	else \
		HEAD = ENTRY; \
	TAIL = ENTRY;

#define _MALLOC_OBJ(OBJ,SIZE) \
	OBJ = malloc(SIZE); \
	if(OBJ == NULL) { \
		print_error_nomem(); \
		exit(ENOMEM); \
	} \
	memset(OBJ,0,SIZE);
//For debugging pre-compiler adds all commands it processed into /tmp/prefile.
//this needs to be removed after replayfs gets matured
extern FILE *prefp;
//Data Buffers List
typedef struct data_entry_s
{
	md5_byte_t hash[MD5_CKSUM_SIZE];
	int len;
	int rat_loc;
	int thread_loc;
	int data_offset;
	timeval_t time;
	struct data_entry_s* next;
} data_entry_t;

typedef struct prog_entry_s prog_entry_t;

//RAT list
typedef struct rat_entry_s
{
	struct res_alloc_table rat;
	struct rat_entry_s* next;
} rat_entry_t;

//For Freeing memory buffers
typedef struct rat_ref_s
{
	int rat_loc;
	struct replay_command* rcmdp;
	struct rat_ref_s* next;
} rat_ref_t;

//For reusing rat entries, resuse_rat_entry() adds, append_rat() removes
typedef struct rat_free_s
{
	int loc;
	struct rat_free_s* next;
} rat_free_t;

//For reusing rat entries, keeps track of FILE->DENTRY->INODE
typedef struct rat_reuse_fid_s
{
	unsigned long rat_loc; //Rat Location
	unsigned long count; //Ref. Count
	union lower_layer {
		long inode;
		long dentry;
		long file;
	}u;
} rat_reuse_fid_t;

//Generic List
typedef struct list_s {
	struct list_s* next;
	void* value;
} list_t;

/*
 * Structures for Oracle Memory Buffer Management
 */
typedef struct mm_oracle_data_s {
	prog_entry_t* pre_prog;
	prog_entry_t* post_prog;
	int pos;
	struct mm_oracle_data_s* next;
} mm_oracle_data_t;

typedef struct mm_oracle_request_s {
	mm_oracle_data_t* data_list;
	mm_oracle_data_t* data_list_tail;
	int buffer_size;
	long data_offset;
	int rat_loc;
	int thread_loc;
	md5_byte_t hash[MD5_CKSUM_SIZE];
} mm_oracle_request_t;

typedef struct mm_oracle_request_node_s {
	md5_byte_t hash[MD5_CKSUM_SIZE];
	int buffer_size;
	long data_offset;
	int rat_loc;
	int thread_loc;
	struct mm_oracle_request_node_s* next;
} mm_oracle_request_node_t;


/* Static Functions */
static int _init_rat();
static int _destroy_rat();
static int _append_new_rat(int val,short type);
static rat_entry_t* _get_rat(int loc);
//RAT is committed by _commit_prog()

static int _commit_prog();
static prog_entry_t* _find_prog(struct replay_command* rcmdp);
static int _finalize_prog(int fd);
static int _destroy_prog();

static int _init_data();
static int _destroy_data();

static int _optimize(); //Garbage Collection, etc
static int _optimize_rat(); //Optimizes the RAT's memory usage.
static int _add_unload_ops_to_last_mem_ref(); //For Buffers

/*
 * LRU Buffer Management - These functions implement LRU Buffer Management
 * The _mm_lru_unload_buffers() function will manage the memory buffer size by
 * adding UNLOAD commands to a queue to be freed before the next VFS
 * operation.	It is called each time a load op to load a memory buffer
 * is added.
 */
static int _mm_lru_init();
static int _mm_lru_destroy();
static int _mm_lru_start_trace_message(op_t op,void* cont);
static int _mm_lru_end_trace_message(op_t op,void* cont);
static int _mm_lru_load_request(long data_offset,long rat_loc,long thread_loc,data_entry_t* entry);
static int _mm_lru_reload_request(md5_byte_t* hash,long data_offset,int buf_len,int rat_loc,int thread_loc);
static int _mm_lru_unload_buffers(long new_buf_size);
//This function will flush the queue to add UNLOAD ops to the PROG
static int _mm_lru_flush_buffers_queue();
//Adds Unload ops for the remaining buffers
static int _mm_lru_unload_remaining_buffers();

/*
 * ORACLE Memory Buffer Management - These functions implement ORACLE_MM
 * Since we are able to determine when a memory buffer is used throughout the
 * replay process, we know the future!	After a buffer is used for the last
 * time, it is automatically deallocated.	At high memory usage, the first
 * buffers deallocated are the ones that are the furthest from the current
 * operation.
 */
static int _mm_oracle_init();
static int _mm_oracle_destroy();
static int _mm_oracle_flush_request_cache();
static int _mm_oracle_start_trace_message(op_t op,void* cont);
static int _mm_oracle_end_trace_message(op_t op,void* cont);
static int _mm_oracle_load_request(long data_offset,long rat_loc,long thread_loc,data_entry_t* entry);
static int _mm_oracle_reload_request(md5_byte_t* hash,long data_offset,int buf_len,int rat_loc,int thread_loc);
static int _mm_oracle_fill_load_cmd(struct replay_command* rcmdp,timeval_t time,long data_offset,int len,int rat_loc,int thread_loc);
static int _mm_oracle_add_load_op(prog_entry_t* prog,long data_offset,int len,int rat_loc,int thread_loc);
static int _mm_oracle_add_unload_op(prog_entry_t* prog,long data_offset,int rat_loc,int thread_loc);
static int _mm_oracle_compare_request(void* key1,int keylen1,
							void* key2,int keylen2);
static int _mm_oracle_process_memory_request_helper(hashed_heap_t* heap,mm_oracle_request_t* request);
static int _mm_oracle_process_memory_requests();
static int _mm_oracle_unload_remaining_buffers();

//For resolving dentry from inode number - for replay_writepage()
static int _add_to_inum_to_dentry_hash(unsigned long inum,unsigned long dentry_loc);
unsigned long _get_dentry_loc_from_inum(unsigned long inum);

//For Reusing RAT entries
#define IS_REUSABLE(TYPE) (TYPE == ARG_INUM || \
				 TYPE == ARG_DENTRY || \
				 TYPE == ARG_FILE)
static int _reuse_rat_obj(int rat_loc);
static int _reuse_generic_get(hash_t* hash,long obj,long lower_obj,unsigned long rat_loc,int type);
static int _reuse_generic_put(hash_t* hash,long obj,int type);
static int _reuse_generic_inc(hash_t* hash,long obj);
inline int _reuse_fget(long file,long dentry,unsigned long rat_loc);
inline int _reuse_fput(long file);
//For Reusing Dentries
inline int _reuse_dget(long dentry,unsigned long inum,unsigned long rat_loc);
inline int _reuse_dput(long dentry);
//For Reusing Inodes
inline int _reuse_iget(unsigned long inode,unsigned long rat_loc);
inline int _reuse_iput(unsigned long inode);

static int _add_load_op(long data_offset,long len,timeval_t time,long rat_loc,long thread_op);
static int _add_unload_op(timeval_t time,long rat_loc,long thread_loc);
static int _request_load_op(long data_offset,long rat_loc,long thread_loc,data_entry_t* entry);
static int _request_reload(md5_byte_t* hash,long data_offset,int buf_len,int rat_loc,int thread_loc);
static int _request_unload_op(timeval_t time,long rat_loc);
static int _unload_remaining_buffers();

//Flush dputs
int _add_remaining_dputs();

//Add create thread op
struct replay_command* _add_create_thread_op(int pid);


/* Static Variables */
static unsigned int rat_list_size;
static rat_entry_t* rat_list_head;
static rat_entry_t* rat_list_tail;

static rat_ref_t* rat_ref_list_head;
static rat_ref_t* rat_ref_list_tail;

static rat_free_t* rat_free_list_head;
static rat_free_t* rat_free_list_tail;

static unsigned int prog_list_size;
static prog_entry_t* prog_list_head;
static prog_entry_t* prog_list_tail;

static int data_fd;
static int data_offset;
static data_entry_t* data_list_head;
static data_entry_t* data_list_tail;
//Hashes MD5 digests to rat locations
#define DATA_LIST_HASH_SIZE 4096
static hash_t data_list_hash;

//LRU-Based Memory Buffer Management
//Hashes MD5 digests to data objs that has been deallocated with OP_UNLOAD
#define MM_LRU_DATA_UNLOADED_HASH_SIZE 4096
static hash_t mm_lru_data_unloaded_hash;
#define MM_LRU_DATA_LIST_HASH_SIZE 4096
static hash_list_t mm_lru_data_list_hash;
//Delayed free queue for memory buffers
static list_t* data_lru_unload_queue;

//ORACLE-Based Memory Buffer Management
static hash_list_t mm_oracle_buffer_requests;
static mm_oracle_request_node_t* mm_oracle_buffer_request_cache;
static mm_oracle_request_node_t* mm_oracle_buffer_request_cache_tail;
static hashed_heap_t mm_oracle_allocated_buffers_heap;
static hash_t mm_oracle_request_data_hash;
static int mm_oracle_counter;
static prog_entry_t* mm_oracle_pre_prog;

//Hashes inode numbers to dentries - for looking up dentries from inode (writepage,syncpage)
//Increased size to 4096 since a larger workload requires this size
#define INUM_TO_DENTRY_HASH_SIZE 4096
static hash_t inum_to_dentry_hash;

//Tracks FILE->DENTRY->INODE RAT objects so that they can be reused
//REUSE_RAT_FID - File
static hash_t reuse_rat_fid_file_hash;
//REUSE_RAT_FID - Dentry
static hash_t reuse_rat_fid_dentry_hash;
//REUSE_RAT_FID - Inode
static hash_t reuse_rat_fid_inode_hash;

//For keeping track of when DGET ops needs to be added
static hash_t unref_dentry_hash; //So we know when to call dget
#define UNREF_DENTRY_HASH_SIZE 4096

//For hashing thread location to dentry location - used during _add_remaining_dputs()
static hash_t dget_op_to_thread_hash;

//For keeping track of DGETs so we can DPUT them later
static hash_list_t dget_hash_list;
int debug_dget_count;

//Hashes RAT locations to RAT pointers
#define RAT_LOCATION_HASH_SIZE 8192
static hash_t rat_location_hash;

//Hashes RAT Locations to RAT Reference objects
static hash_t rat_ref_hash;

//Hashes rat values and types to rat locations
#define RAT_HASH_SIZE 8192
static hash_t rat_hash;

//Watermark for memory buffers - in megabytes
static int memory_buffers_max_size;
static long memory_buffers_size; //Current Size

static char* filename; //Path to RAT and PROG files


int res_manager_init(char* fname,int memory_buffers_size)
{
	print_entry_location();

	filename = malloc(PATH_MAX);
	memory_buffers_max_size = memory_buffers_size;
	if(!filename) {
		print_error_nomem();
		exit(ENOMEM);
	}

	strncpy(filename,fname,PATH_MAX);

	_init_rat();
	res_manager_init_prog();
	_init_data();

	//Init Hashes
	hash_init(&rat_hash,RAT_HASH_SIZE);
	hash_init(&rat_location_hash,RAT_LOCATION_HASH_SIZE);
	hash_init(&rat_ref_hash,0);
	hash_init(&inum_to_dentry_hash,INUM_TO_DENTRY_HASH_SIZE);
	hash_init(&reuse_rat_fid_file_hash,0);
	hash_init(&unref_dentry_hash,0);
	hash_init(&dget_op_to_thread_hash,0);
	hash_list_init(&dget_hash_list,0);
	//hash_init(&reuse_rat_fid_dentry_hash,0);
	//hash_init(&reuse_rat_fid_inode_hash,0);

	//Init Hash lists
	if(!ORACLE_MM)
		_mm_lru_init();
	else
		_mm_oracle_init();

#if MULTITHREADING
	thread_manager_init();
#endif

	print_exit_location();
	return 0;
}

//Called whenever a new VFS op is being processed
int res_manager_start_trace_message(op_t op,void* cont)
{
	print_entry_location();

	if(ORACLE_MM)
		_mm_oracle_start_trace_message(op,cont);
	else
		_mm_lru_start_trace_message(op,cont);

	print_exit_location();
	return 0;
}

//Called whenever a new VFS op is being processed
int res_manager_end_trace_message(op_t op,void* cont)
{
	print_entry_location();

	if(ORACLE_MM)
		_mm_oracle_end_trace_message(op,cont);
	else
		_mm_lru_end_trace_message(op,cont);

	print_exit_location();
	return 0;
}

int res_manager_commit()
{
	print_entry_location();

	//If using Oracle MM then process memory requests
	if(ORACLE_MM)
		_mm_oracle_process_memory_requests();

#if MULTITHREADING
	thread_manager_process_thread_requests();
#endif

	//Optimize
	_optimize();

	if(ORACLE_MM)
		_mm_oracle_destroy();
	else
		_mm_lru_destroy();

#if MULTITHREADING
	thread_manager_destroy();
#endif

	if(prog_list_tail)
		_add_remaining_dputs();

	//Commit Program
	_commit_prog();

	//Cleanup
	_destroy_rat();
	_destroy_prog();
	_destroy_data();

	//Destroy Hashes and hashed objects
	hash_destroy(&rat_hash);
	hash_destroy(&rat_location_hash);
	hash_destroy(&inum_to_dentry_hash);
	hash_destroy_all(&reuse_rat_fid_file_hash);
	hash_destroy(&unref_dentry_hash);
	hash_destroy(&dget_op_to_thread_hash);
	hash_list_destroy(&dget_hash_list);
	//hash_destroy_all(&reuse_rat_fid_dentry_hash);
	//hash_dentroy_all(&reuse_rat_fid_inode_hash);

	free(filename);

	print_exit_location();
	return 0;
}

static int _init_rat()
{
	print_entry_location();

	rat_list_size = MIN_RAT_SIZE;
	rat_list_head = NULL;
	rat_list_tail = NULL;

	print_exit_location();
	return 0;
}

int res_manager_lookup_rat(int val, short type)
{
	int ret;
	int keysize = sizeof(val) + sizeof(type);
	char key[keysize];
	print_entry_location();

	memcpy(key,&val,sizeof(val));
	memcpy(key+sizeof(val),&type,sizeof(type));

	ret = (int)hash_find(&rat_hash,key,keysize);
	if(!ret)
		ret = -1;

	print_exit_location();
	return ret;
}

static rat_entry_t* _get_rat(int loc)
{
	rat_entry_t* rptr;
	print_entry_location();

	rptr = hash_find(&rat_location_hash,(void*)&loc,sizeof(loc));

	print_exit_location();
	return rptr;
}

int res_manager_append_rat(int val,short type)
{
	long loc = 0;
	rat_entry_t* rptr;
	rat_free_t* rat_free;
	int keysize = sizeof(val) + sizeof(type);
	char key[keysize];

	print_entry_location();

	if(!rat_free_list_head || !IS_REUSABLE(type)) {
		loc = _append_new_rat(val,type);
		goto out;
	}

	//Reuse RAT entry
	loc = rat_free_list_head->loc;
	rat_free = rat_free_list_head;
	rat_free_list_head = rat_free_list_head->next;
	free(rat_free);

	rptr = _get_rat(loc);
	if(rptr == NULL) {
		print_warning_message("get_rat(%d): FAILED!",(int)loc);
		loc = _append_new_rat(val,type);
		goto out;
	}
	rptr->rat.type = type;
	rptr->rat.u.ival = val;

	//Add to Rat Hash
	memcpy(key,&val,sizeof(val));
	memcpy(key+sizeof(val),&type,sizeof(type));

	if(type == ARG_STR)
		assert(!hash_add_allow_dupes(&rat_hash,key,keysize,(void*)loc));
	else
		assert(!hash_add(&rat_hash,key,keysize,(void*)loc));

 out:
	print_exit_location();
	return loc;
}

int _append_new_rat(int val, short type)
{
	int loc;
	int keysize = sizeof(val) + sizeof(type);
	char key[keysize];
	rat_entry_t* entry;
	print_entry_location();

	_ADD_TO_LIST(rat_list_head,rat_list_tail,rat_list_tail->next,entry,sizeof(rat_entry_t));

	loc = rat_list_size++;
	hash_add_replace_dupes(&rat_location_hash,(void*)&loc,sizeof(loc),entry);

	entry->rat.type = type;
	entry->rat.u.ival = val;

	//Add to Rat Hash
	memcpy(key,&val,sizeof(val));
	memcpy(key+sizeof(val),&type,sizeof(type));

	if(type == ARG_STR)
		assert(!hash_add_allow_dupes(&rat_hash,key,keysize,(void*)loc));
	else
		assert(!hash_add(&rat_hash,key,keysize,(void*)loc));

	print_exit_location();
	return loc;
}

int res_manager_add_rat_ref(struct replay_command* rcmdp,int rat_loc)
{
	rat_ref_t* entry;
	print_entry_location();

	assert(rcmdp != NULL);

	entry = hash_find(&rat_ref_hash,(void*)&rat_loc,sizeof(rat_loc));
	if(entry != NULL) {
		goto out;
		entry->rcmdp = rcmdp;
	}

	_ADD_TO_LIST(rat_ref_list_head,rat_ref_list_tail,rat_ref_list_tail->next,entry,sizeof(rat_ref_t));

	entry->rcmdp = rcmdp;
	entry->rat_loc = rat_loc;
	hash_add(&rat_ref_hash,(void*)&rat_loc,sizeof(rat_loc),entry);

 out:
	print_exit_location();
	return 0;
}

static int _add_to_inum_to_dentry_hash(unsigned long inum,unsigned long dentry_loc)
{
	int ret = -1;
	print_entry_location();
	assert(dentry_loc >= 0);

	ret = (int)hash_find(&inum_to_dentry_hash,&inum,sizeof(inum));
	//If the mapping is already found in the hash, first remove the mapping and later add
	//another one since it results in conflicts
	if(ret)
	{
		fprintf(prefp,"inum already in hash, removing...");
		hash_remove(&inum_to_dentry_hash,&inum, sizeof(inum));
	}
	fprintf(prefp," added ino %ld : dloc %ld",inum,dentry_loc);
	hash_add(&inum_to_dentry_hash,(void*)&inum,sizeof(inum),(void*)dentry_loc);

	print_exit_location();
	return 0;
}

unsigned long _get_dentry_loc_from_inum(unsigned long inum)
{
	int ret = -1;
	print_entry_location();

	ret = (int)hash_find(&inum_to_dentry_hash,&inum,sizeof(inum));

	print_exit_location();
	return ret;
}

int res_manager_resolve_dentry_from_inum(struct context* cont,int op,int pre_or_post,struct replay_command** rcmd_pre,struct replay_command** rcmd_post)
{
	int dentry_loc;
	print_entry_location();
	assert(cont != NULL);

	if(!pre_or_post)
		goto out;

	switch(op) {
	case OP_LOOKUP:
		//Only if its a valid inode number add it to hash
		//the dentry loc of is stored in args[2] in trfs_pre.c. changed from args[0] -> args[2]
		if(cont->saved[CONTEXT_SAVED_INUM])
			_add_to_inum_to_dentry_hash(cont->saved[CONTEXT_SAVED_INUM],cont->args[2]);
		break;
	//Added three more cases to add the inode-dentry mapping when nodes/ files/ directories
	//are created
	case OP_CREATE:
	case OP_MKDIR:
	case OP_MKNOD:
        _add_to_inum_to_dentry_hash(cont->args[5],cont->args[0]);
        break;

	case OP_PERMISSION:
	case OP_SYNC_PAGE:
	case OP_WRITEPAGE:
		dentry_loc = _get_dentry_loc_from_inum(cont->saved[CONTEXT_SAVED_INUM]);
		assert(dentry_loc != RAT_ROOT_DENTRY); //No address_space ops for directories
		if(dentry_loc)
		{
			//Found
			cont->args[0] = dentry_loc;
			if(res_manager_remove_unref_dentry(cont->ret))
			{
				//Add DGET
				res_manager_add_dget_op(rcmd_post,cont->time,dentry_loc,cont->pid);
			}
			fprintf(prefp,"\nOP: %d Found dentry loc %d for u=ino %d",op,cont->args[0],cont->saved[CONTEXT_SAVED_INUM]);
			goto out;
		}
		//Not Found
		//XXX: Tim: Assuming the inode's dentry is the ROOT DENTRY if not in hash
		if(op == OP_WRITEPAGE || op == OP_SYNC_PAGE)
			fprintf(prefp,"\nOP: %d No Dentry for %d: Assumin ROOT_DNTRY",op,cont->saved[CONTEXT_SAVED_INUM]);
		cont->args[0] = RAT_ROOT_DENTRY;
		break;
	}

 out:
	print_exit_location();
	return 0;
}

int res_manager_add_dget_op(struct replay_command** cmd,timeval_t time,long rat_loc,long thread_loc)
{
	long* loc;
	print_entry_location();
	//Make sure it's not already being used or allocated
	assert(*cmd == NULL);

	loc = malloc(sizeof(long));
	if(!loc) {
		print_error_nomem();
		exit(ENOMEM);
	}
	*loc = rat_loc;

	hash_list_add_last(&dget_hash_list,&rat_loc,sizeof(rat_loc),loc);
	*cmd = malloc(sizeof(struct replay_command));
	if(*cmd == NULL) {
		print_error_nomem();
		exit(ENOMEM);
	}
	memset(*cmd,0,sizeof(struct replay_command));

	(*cmd)->command = OP_DGET;
	(*cmd)->expected_retval = 0;
	(*cmd)->args[0] = rat_loc;
	(*cmd)->time = time;
	(*cmd)->thread_loc = thread_loc;
	(*cmd)->ret = 0;

	res_manager_hash_dget_op_to_thread(rat_loc,thread_loc);

	print_exit_location();
	return 0;
}

int res_manager_add_dput_op(struct replay_command** cmd,timeval_t time,long rat_loc,long thread_loc)
{
	long *loc;
	print_entry_location();

	//Make sure it's not already being used or allocated
	assert(*cmd == NULL);

	loc = hash_remove(&dget_op_to_thread_hash,&rat_loc,sizeof(rat_loc));
	free(loc); //Throw out
	loc = (long*)hash_list_remove(&dget_hash_list,&rat_loc,sizeof(rat_loc));
	assert(loc);
	free(loc);
	*cmd = malloc(sizeof(struct replay_command));
	if(*cmd == NULL) {
		print_error_nomem();
		exit(ENOMEM);
	}
	memset(*cmd,0,sizeof(struct replay_command));

	(*cmd)->command = OP_DPUT;
	(*cmd)->expected_retval = 0;
	(*cmd)->args[0] = rat_loc;
	(*cmd)->time = time;
	(*cmd)->thread_loc = 0;
	(*cmd)->ret = 0;

	print_exit_location();
	return 0;
}

int _add_remaining_dputs()
{
	prog_entry_t* entry;
	long *loc;
	long rat_loc;
	int *thread_loc;
	timeval_t time;
	print_entry_location();

	time = prog_list_tail->rcmdp->time;
	loc = (long*)hash_list_get_first(&dget_hash_list);
	while(loc) {
		rat_loc = *loc;
		_ADD_TO_LIST(prog_list_head,prog_list_tail,prog_list_tail->next,entry,sizeof(prog_entry_t));
		prog_list_size++;

		//Find thread location
		thread_loc = hash_remove(&dget_op_to_thread_hash,loc,sizeof(*loc));
		assert(thread_loc);
		res_manager_add_dput_op(&entry->rcmdp,time,rat_loc,*thread_loc);
		free(thread_loc);
		loc = (long*)hash_list_get_first(&dget_hash_list);
	}

	print_exit_location();
	return 0;
}

static int _add_load_op(long data_offset,long len,timeval_t time,long rat_loc,long thread_loc)
{
	struct replay_command rcmd;
	print_entry_location();

	memset(&rcmd,0,sizeof(struct replay_command));
	rcmd.command = OP_LOAD;
	rcmd.args[0] = data_offset;
	rcmd.args[1] = len;
	rcmd.time = time;
	rcmd.thread_loc = thread_loc;
	rcmd.ret = rat_loc;

	res_manager_append_prog(&rcmd);

	print_exit_location();
	return 0;
}

static int _add_unload_op(timeval_t time,long rat_loc,long thread_loc)
{
	struct replay_command rcmd;
	list_t* entry;
	print_entry_location();

	memset(&rcmd,0,sizeof(struct replay_command));
	rcmd.command = OP_UNLOAD;
	rcmd.args[0] = rat_loc;
	rcmd.time = time;
	rcmd.thread_loc = thread_loc;
	rcmd.ret = 0;

	res_manager_append_prog(&rcmd);

	print_exit_location();
	return 0;
}

static int _request_load_op(long data_offset,long rat_loc,long thread_loc,data_entry_t* entry)
{
	int err = 0;
	print_entry_location();

	if(ORACLE_MM)
		err = _mm_oracle_load_request(data_offset,rat_loc,thread_loc,entry);
	else
		err = _mm_lru_load_request(data_offset,rat_loc,thread_loc,entry);

	print_exit_location();
	return err;
}

static int _request_reload(md5_byte_t* hash,long data_offset,int buf_len,int rat_loc,int thread_loc)
{
	int err = 0;
	print_entry_location();

	if(ORACLE_MM)
		err = _mm_oracle_reload_request(hash,data_offset,buf_len,rat_loc,thread_loc);
	else
		err = _mm_lru_reload_request(hash,data_offset,buf_len,rat_loc,thread_loc);

	print_exit_location();
	return err;
}

static int _unload_remaining_buffers()
{
	int err = 0;
	print_entry_location();

	if(ORACLE_MM)
		err = _mm_oracle_unload_remaining_buffers();
	else
		err = _mm_lru_unload_remaining_buffers();

	print_exit_location();
	return err;
}

static int _mm_lru_init()
{
	int err = 0;
	print_entry_location();

	hash_list_init(&mm_lru_data_list_hash,MM_LRU_DATA_LIST_HASH_SIZE);
	data_lru_unload_queue = NULL;

	print_exit_location();
	return err;
}

static int _mm_lru_destroy()
{
	//Destroy Hash Lists
	hash_list_destroy(&mm_lru_data_list_hash);

	//Add UNLOAD ops for the remaining memory buffers
	_mm_lru_unload_remaining_buffers();

	return 0;
}

static int _mm_lru_start_trace_message(op_t op,void* cont)
{
	int err = 0;
	print_entry_location();

	err = _mm_lru_flush_buffers_queue();

	print_exit_location();
	return err;
}

static int _mm_lru_end_trace_message(op_t op,void* cont)
{
	int err = 0;
	print_entry_location();

	//Do Nothing

	print_exit_location();
	return err;
}

static int _mm_lru_load_request(long data_offset,long rat_loc,long thread_loc,data_entry_t* entry)
{
	int err = 0;
	print_entry_location();
	assert(entry);

	assert(!_add_load_op(data_offset,entry->len,entry->time,rat_loc,thread_loc));

	//Add UNLOAD ops to limit total memory buffer allocation
	_mm_lru_unload_buffers(entry->len);

	//Add the data entry onto the LRU list
	assert(!hash_list_add_first(&mm_lru_data_list_hash,&rat_loc,sizeof(rat_loc),entry));
	err = rat_loc;

	print_exit_location();
	return err;
}

static int _mm_lru_reload_request(md5_byte_t* hash,long data_offset,int buf_len,int rat_loc,int thread_loc)
{
	int err = 0;
	data_entry_t* ptr;
	print_entry_location();

	//Check to see if this buffer has been processed and deallocated
	ptr = hash_remove(&mm_lru_data_unloaded_hash,(void*)hash,MD5_CKSUM_SIZE);
	if(ptr) {
		//Add LOAD op
		_add_load_op(ptr->data_offset,ptr->len,ptr->time,ptr->rat_loc,ptr->thread_loc);

		//Add UNLOAD ops to limit total memory buffer allocation
		_mm_lru_unload_buffers(buf_len);

		assert(!hash_list_add_first(&mm_lru_data_list_hash,&ptr->rat_loc,sizeof(ptr->rat_loc),ptr));
		err = ptr->rat_loc;
		goto out;
	}

	//"Touch" the data entry on the LRU list if it's in the LRU
	ptr = hash_list_remove(&mm_lru_data_list_hash,&rat_loc,sizeof(rat_loc));
	if(ptr)
		assert(!hash_list_add_first(&mm_lru_data_list_hash,&rat_loc,sizeof(rat_loc),ptr));

 out:
	print_exit_location();
	return err;
}

/*
 * Called whenever a new memory buffer is loaded, this will add UNLOAD ops to
 * adjust memory buffer size
 */
static int _mm_lru_unload_buffers(long new_buf_size)
{
	int err = 0;
	data_entry_t* entry;
	list_t* lptr;
	list_t* new_list;
	print_entry_location();

	memory_buffers_size += new_buf_size;
	if(memory_buffers_size <= (memory_buffers_max_size * 1000000))
		goto out;

	//Goto the last element in the free queue
	lptr = data_lru_unload_queue;
	while(lptr != NULL && lptr->next != NULL)
		lptr = lptr->next;

	//Queue Frees until it is beneath the watermark
	while(memory_buffers_size > (memory_buffers_max_size * 1000000)) {
		entry = (data_entry_t*)hash_list_remove_last(&mm_lru_data_list_hash);
		assert(entry != NULL);

		//Add to Data Deallocate Queue
		_MALLOC_OBJ(new_list,sizeof(list_t));
		new_list->value = entry;
		if(data_lru_unload_queue == NULL)
			data_lru_unload_queue = new_list;
		else
			lptr->next = new_list;

		lptr = new_list;
		memory_buffers_size -= entry->len;
	}

 out:
	print_exit_location();
	return err;
}

static int _mm_lru_flush_buffers_queue()
{
	list_t* lptr;
	list_t* prev;
	data_entry_t* entry;
	int digest_len = MD5_CKSUM_SIZE;
	print_entry_location();

	lptr = data_lru_unload_queue;
	while(lptr != NULL) {
		entry = (data_entry_t*)lptr->value;
		assert(entry != NULL);

		//Add unload op
		_add_unload_op(entry->time,entry->rat_loc,entry->thread_loc);

		//Add to the deallocated_hash
		assert(!hash_add(&mm_lru_data_unloaded_hash,(void*)entry->hash,digest_len,entry));

		prev = lptr;
		lptr = lptr->next;
		free(prev);
	}

	data_lru_unload_queue = NULL;

	print_exit_location();
	return 0;
}

static int _mm_lru_unload_remaining_buffers()
{
	data_entry_t* dptr;
	data_entry_t* entry;
	print_entry_location();

	dptr = data_list_head;
	while(dptr != NULL) {
		entry = hash_remove(&mm_lru_data_unloaded_hash,(void*)dptr->hash,dptr->len);
		if(entry != NULL)
			continue;

		//Add Unload Op
		_add_unload_op(entry->time,entry->rat_loc,entry->thread_loc);
		dptr = dptr->next;
	}

	print_exit_location();
	return 0;
}

static int _mm_oracle_init()
{
	int err = 0;
	print_entry_location();

	hash_list_init(&mm_oracle_buffer_requests,0);
	hashed_heap_init(&mm_oracle_allocated_buffers_heap,0,_mm_oracle_compare_request);
	hash_init(&mm_oracle_request_data_hash,0);
	mm_oracle_buffer_request_cache = NULL;
	mm_oracle_buffer_request_cache_tail = NULL;
	mm_oracle_counter = 0;

	print_exit_location();
	return err;
}

static int _mm_oracle_destroy()
{
	hash_list_destroy(&mm_oracle_buffer_requests);
	hashed_heap_destroy(&mm_oracle_allocated_buffers_heap);
	hash_destroy(&mm_oracle_request_data_hash);

	return 0;
}

static int _mm_oracle_start_trace_message(op_t op,void* cont)
{
	int err = 0;
	print_entry_location();

	mm_oracle_pre_prog = prog_list_tail;

	print_exit_location();
	return err;
}

//Flush cached memory request to request queue and attach counter and prog
static int _mm_oracle_flush_request_cache()
{
	int err = 0;
	mm_oracle_data_t* data;
	mm_oracle_request_t* request;
	mm_oracle_request_node_t* cache = mm_oracle_buffer_request_cache;
	mm_oracle_request_node_t* prev_cache;
	print_entry_location();

	if(!cache)
		goto out;

	//Flush Request Cache to Request Queue and attach data
	while(cache) {
		//Allocate and fill data object
		data = malloc(sizeof(mm_oracle_data_t));
		if(!data) {
			print_error_nomem();
			exit(ENOMEM);
		}
		memset(data,0,sizeof(mm_oracle_data_t));

		data->pos = mm_oracle_counter;
		data->pre_prog = mm_oracle_pre_prog;
		data->post_prog = prog_list_tail;

		request = hash_list_find(&mm_oracle_buffer_requests,cache->hash,MD5_CKSUM_SIZE);

		//Allocate New Request Entry
		if(!request) {
			request = malloc(sizeof(mm_oracle_request_t));
			if(!request) {
	print_error_nomem();
	exit(ENOMEM);
			}
			memset(request,0,sizeof(mm_oracle_request_t));

			//Fill Request Object
			request->buffer_size = cache->buffer_size;
			request->rat_loc = cache->rat_loc;
			request->thread_loc = cache->thread_loc;
			request->data_offset = cache->data_offset;
			memcpy(request->hash,cache->hash,MD5_CKSUM_SIZE);

			//Make new list
			request->data_list = request->data_list_tail = data;
		}
		else {
			//Append to list
			request->data_list_tail->next = data;
			request->data_list_tail = data;
		}

		//Append to Request Queue
		hash_list_add_last(&mm_oracle_buffer_requests,request->hash,MD5_CKSUM_SIZE,request);

		prev_cache = cache;
		cache = cache->next;
		free(prev_cache);
	}

	mm_oracle_buffer_request_cache = NULL;
	mm_oracle_buffer_request_cache_tail = NULL;
	mm_oracle_counter++;

 out:
	print_exit_location();
	return err;
}

static int _mm_oracle_end_trace_message(op_t op,void* cont)
{
	return _mm_oracle_flush_request_cache();
}

static int _mm_oracle_load_request(long data_offset,long rat_loc,long thread_loc,data_entry_t* entry)
{
	int err = 0;
	print_entry_location();
	assert(entry);

	err = _mm_oracle_reload_request(entry->hash,data_offset,entry->len,rat_loc,thread_loc);

	print_exit_location();
	return err;
}

static int _mm_oracle_reload_request(md5_byte_t* hash,long data_offset,int buf_len,int rat_loc,int thread_loc)
{
	int err = 0;
	mm_oracle_request_node_t* node;
	print_entry_location();
	assert(hash);
	assert(buf_len > 0 && rat_loc > 0);

	//Setup Node
	node = malloc(sizeof(mm_oracle_request_node_t));
	if(!node) {
		print_error_nomem();
		exit(ENOMEM);
	}
	memset(node,0,sizeof(mm_oracle_request_node_t));
	memcpy(node->hash,hash,MD5_CKSUM_SIZE);
	node->buffer_size = buf_len;
	node->data_offset = data_offset;
	node->rat_loc = rat_loc;
	node->thread_loc = thread_loc;

	//Append to request cache list
	if(mm_oracle_buffer_request_cache) {
		mm_oracle_buffer_request_cache_tail->next = node;
		mm_oracle_buffer_request_cache_tail = node;
	}
	else {
		mm_oracle_buffer_request_cache = node;
		mm_oracle_buffer_request_cache_tail = node;
	}

	print_exit_location();
	return err;
}

static int _mm_oracle_fill_load_cmd(struct replay_command* rcmdp,timeval_t time,long data_offset,int len,int rat_loc,int thread_loc)
{
	int err = 0;
	print_entry_location();
	assert(rcmdp);

	memset(rcmdp,0,sizeof(struct replay_command));
	rcmdp->command = OP_LOAD;
	rcmdp->args[0] = data_offset;
	rcmdp->args[1] = len;
	rcmdp->time = time;
	rcmdp->thread_loc = thread_loc;
	rcmdp->ret = rat_loc;

	print_exit_location();
	return err;
}

static int _mm_oracle_add_load_op(prog_entry_t* prog,long data_offset,int len,int rat_loc,int thread_loc)
{
	int err = 0;
	struct replay_command rcmd;
	print_entry_location();

	_mm_oracle_fill_load_cmd(&rcmd,prog->rcmdp->time,data_offset,len,rat_loc,thread_loc);
	res_manager_insert_prog(prog,&rcmd);

	print_exit_location();
	return err;
}

static int _mm_oracle_add_unload_op(prog_entry_t* prog,long data_offset,int rat_loc,int thread_loc)
{
	int err = 0;
	struct replay_command rcmd;
	print_entry_location();
	assert(prog);

	memset(&rcmd,0,sizeof(struct replay_command));
	rcmd.command = OP_UNLOAD;
	rcmd.args[0] = rat_loc;
	rcmd.args[1] = data_offset;
	rcmd.thread_loc = thread_loc;
	rcmd.time = prog->rcmdp->time;
	rcmd.ret = 0;

	res_manager_insert_prog(prog,&rcmd);

	print_exit_location();
	return err;
}

static int _mm_oracle_compare_request(void* key1,int keylen1,
							void* key2,int keylen2)
{
	int err = 0;
	mm_oracle_request_t* request1;
	mm_oracle_request_t* request2;
	int value1,value2;
	print_entry_location();
	assert(key1 && keylen1 > 0);
	assert(key2 && keylen2 > 0);

	request1 = hash_list_find(&mm_oracle_buffer_requests,key1,keylen1);
	request2 = hash_list_find(&mm_oracle_buffer_requests,key2,keylen2);

	//BUG if not found
	assert(request1);
	assert(request2);

	//BUG if data_list is empty
	assert(request1->data_list);
	assert(request2->data_list);

	value1 = request1->data_list->pos;
	value2 = request2->data_list->pos;

	if(value1 == value2)
		err = 0;
	else if(value1 > value2)
		err = 1;
	else
		err = -1;

	print_exit_location();
	return err;
}

static int _mm_oracle_process_memory_request_helper(hashed_heap_t* heap,mm_oracle_request_t* request)
{
	int err = 0;
	mm_oracle_request_t* heap_entry;
	mm_oracle_data_t* request_data;
	print_entry_location();
	assert(heap);
	assert(request);

	request_data = request->data_list;
	assert(request_data);

	//Pop Data_list - Shift to next use position
	request->data_list = request->data_list->next;

	if(request->data_list)
		assert(request_data->pos != request->data_list->pos);

	//Is buffer already allocated?
	heap_entry = hashed_heap_find(heap,request->hash,MD5_CKSUM_SIZE);
	assert(request->thread_loc >= 0);
	if(!heap_entry) {
		if(request_data->next)
			hashed_heap_add(heap,request->hash,MD5_CKSUM_SIZE,request);
		if(request_data->pre_prog)
			_mm_oracle_add_load_op(request_data->pre_prog,
					 request->data_offset,
					 request->buffer_size,
					 request->rat_loc,
					 request->thread_loc);
		else {
			//Load needs to be appended to the front of the prog list
			prog_entry_t* entry;
			_MALLOC_OBJ(entry,sizeof(prog_entry_t));
			_MALLOC_OBJ(entry->rcmdp,sizeof(struct replay_command));

			_mm_oracle_fill_load_cmd(entry->rcmdp,
						 request_data->post_prog->rcmdp->time,
						 request->data_offset,
						 request->buffer_size,
						 request->rat_loc,
						 request->thread_loc);
			entry->next = prog_list_head;
			prog_list_head = entry;
			prog_list_size++;
		}
		memory_buffers_size += request->buffer_size;

		hash_add(&mm_oracle_request_data_hash,(void*)&request,sizeof(void*),request_data);
	}
	else {
		mm_oracle_data_t* old_data;
		//Already allocated - Remove off the heap, and re-add with next use position
		assert(hashed_heap_remove(heap,request->hash,MD5_CKSUM_SIZE));
		if(request_data->next)
			assert(!hashed_heap_add(heap,request->hash,MD5_CKSUM_SIZE,request));

		//Remove previous entry in hash and re-add new one
		old_data = hash_remove(&mm_oracle_request_data_hash,
					 (void*)&request,
					 sizeof(void*));
		assert(old_data);
		hash_add(&mm_oracle_request_data_hash,(void*)&request,sizeof(void*),request_data);
		free(old_data);
	}

	//If we are above the allowable memory buffer watermark, add unload ops
	while(memory_buffers_size > (memory_buffers_max_size * 1000000)) {
		mm_oracle_data_t* old_data;
		heap_entry = hashed_heap_remove_root(heap);
		assert(heap_entry);
		old_data = hash_remove(&mm_oracle_request_data_hash,
				 (void*)&heap_entry,
				 sizeof(void*));
		assert(old_data);

		memory_buffers_size -= heap_entry->buffer_size;
		assert(memory_buffers_size >= 0);
		assert(heap_entry->thread_loc >= 0);
		_mm_oracle_add_unload_op(old_data->post_prog,
					 heap_entry->data_offset,
					 heap_entry->rat_loc,
					 heap_entry->thread_loc);
		free(old_data);
	}

	//Free request data and add_unload_op if it's the buffer's last use
	if(!request_data->next) {
		mm_oracle_data_t* data;
		assert(request->thread_loc >= 0);
		//This is the last time the memory buffer is used, so unload it!
		_mm_oracle_add_unload_op(request_data->post_prog,
					 request->data_offset,
					 request->rat_loc,
					 request->thread_loc);
		memory_buffers_size -= request->buffer_size;
		hashed_heap_remove(heap,request->hash,MD5_CKSUM_SIZE);
		data = hash_remove(&mm_oracle_request_data_hash,(void*)&request,sizeof(void*));
		assert(data);
		free(data);
	}

 out:
	print_exit_location();
	return err;
}

/*
 * This is called after all the memory requests have been made.	This will
 * insert ops to load and unload memory buffers.
 */
static int _mm_oracle_process_memory_requests()
{
	int err = 0;
	mm_oracle_request_t* request;
	hashed_heap_t* heap = &mm_oracle_allocated_buffers_heap;
	print_entry_location();

	request = hash_list_get_first(&mm_oracle_buffer_requests);
	while(request) {
		//Process the Memory Request
		_mm_oracle_process_memory_request_helper(heap,request);
		//Fetch next request
		assert(request = hash_list_remove_first(&mm_oracle_buffer_requests));
		if(!request->data_list)
			free(request);
		request = hash_list_get_first(&mm_oracle_buffer_requests);
	}

	_mm_oracle_unload_remaining_buffers();

	print_exit_location();
	return err;
}

static int _mm_oracle_unload_remaining_buffers()
{
	int err = 0;
	print_entry_location();

	/*
	 * Since we free every buffer after its last use there shouldn't be
	 * any buffers left.
	 */
	assert(!hashed_heap_remove_root(&mm_oracle_allocated_buffers_heap));
	assert(memory_buffers_size == 0);

	print_exit_location();
	return err;
}

int res_manager_add_unref_dentry(long rat_loc)
{
	return hash_add(&unref_dentry_hash,(void*)&rat_loc,sizeof(rat_loc),(void*)1);
}

int res_manager_remove_unref_dentry(long rat_loc)
{
	return (int)hash_remove(&unref_dentry_hash,(void*)&rat_loc,sizeof(rat_loc));
}

int res_manager_hash_dget_op_to_thread(long rat_loc,int thread_loc)
{
	long *loc = malloc(sizeof(int));
	if(!loc) {
		print_error_nomem();
		exit(ENOMEM);
	}
	*loc = thread_loc;
	hash_add_allow_dupes(&dget_op_to_thread_hash,&rat_loc,sizeof(rat_loc),loc);
	return 0;
}

//For reusing F->D->I objects
int res_manager_reuse_rat_entries(struct context* cont,int op,int pre_or_post)
{
	print_entry_location();
	assert(cont != NULL);

	if(!pre_or_post) {
		if(op != OP_OPEN && op != OP_RELEASE)
			goto out;
	}
	else {
		if(op == OP_OPEN || op == OP_RELEASE)
			goto out;
	}

	if(cont->saved_fid_flags | CONTEXT_SAVED_FID_DENTRY_GET) {
		/*_reuse_dget(cont->saved[CONTEXT_SAVED_DENTRY_PTR],
				 cont->saved[CONTEXT_SAVED_INUM],
				 cont->saved[CONTEXT_SAVED_RAT_LOC]); */
	}

	switch(op) {
	/* File Ops */
	case OP_OPEN:
		_reuse_fget(cont->saved[CONTEXT_SAVED_FILE_PTR],
				 cont->saved[CONTEXT_SAVED_DENTRY_PTR],
				 cont->saved[CONTEXT_SAVED_RAT_LOC]);
		break;
	case OP_RELEASE:
		_reuse_fput(cont->saved[CONTEXT_SAVED_FILE_PTR]);
		break;
	/* Dentry Ops */
	case OP_D_RELEASE:
		//_reuse_dput(cont->saved[CONTEXT_SAVED_DENTRY_PTR]);
		break;
	/* Inode Ops */
	//These are not implemented since inodes are not currently used in the RAT
	case OP_LOOKUP:
		/*res_manager_iget(cont->saved[CONTEXT_SAVED_INUM],
			cont->saved[CONTEXT_RAT_LOC]); */
		break;
	case OP_D_IPUT:
		//res_manager_iput(cont->saved[CONTEXT_SAVED_INUM]);
	default:
		goto out;
	}

 out:
	print_exit_location();
	return 0;
}

int _reuse_generic_get(hash_t* hash,long obj,long lower_obj,unsigned long rat_loc,int type)
{
	int err = 0;
	rat_reuse_fid_t* entry;
	print_entry_location();

	//See if obj already exists
	entry = hash_find(hash,(void*)&obj,sizeof(obj));
	if(entry != NULL) {
		entry->count++;
		goto out;
	}

	//Fill Object
	_MALLOC_OBJ(entry,sizeof(rat_reuse_fid_t));
	entry->rat_loc = rat_loc;
	entry->count = 1;
	switch(type) {
	case ARG_INUM:
		entry->u.inode = lower_obj;
		break;
	case ARG_DENTRY:
		entry->u.dentry = lower_obj;
		/*
		 * Inodes not currently used in the RAT
		 */
		//_reuse_generic_inc(&reuse_rat_fid_inode_hash,lower_obj);
		break;
	case ARG_FILE:
		entry->u.file = lower_obj;
		//_reuse_generic_inc(&reuse_rat_fid_dentry_hash,lower_obj);
		break;
	default:
		print_error_message("Invalid Type, cannot 'get' object correctly, removing from reuse list");
		free(entry);
		err = EINVAL;
		goto out;
	}

	//Hash
	hash_add(hash,(void*)&obj,sizeof(obj),entry);

 out:
	print_exit_location();
	return err;
}

//For reusing F->D->I objects
int _reuse_generic_put(hash_t* hash,long obj,int type)
{
	int err = 0;
	rat_reuse_fid_t* entry;
	rat_entry_t* rat;
	print_entry_location();

	entry = hash_find(hash,(void*)&obj,sizeof(obj));

	if(!entry) {
		print_error_message("Unable to 'put' object, object: %p not found",(void*)obj);
		err = -1;
		goto out;
	}

	if(--entry->count <= 0) {
		entry = hash_remove(hash,(void*)&obj,sizeof(obj));

		//Call dput to drop decrement lower layer ref. count
		switch(type) {
		case ARG_INUM:
			break;
		case ARG_DENTRY:
			/*
			 * There are currently no inodes in the RAT
			 */
			//_reuse_iput(entry->u.inode);
			break;
		case ARG_FILE:
			//_reuse_dput(entry->u.dentry);
			break;
		default:
			print_error_message("Invalid Type, cannot 'put' object correctly");
			free(entry);
			err = EINVAL;
			goto out;
		}
		_reuse_rat_obj(entry->rat_loc);
		free(entry);
	}

out:
	print_exit_location();
	return err;
}

int _reuse_generic_inc(hash_t* hash,long obj)
{
	rat_reuse_fid_t* entry;
	print_entry_location();

	assert(hash != NULL);

	entry = hash_find(hash,(void*)&obj,sizeof(obj));
	if(entry == NULL) {
		print_error_message("Object is not found");
		goto out;
	}

	entry->count++;

out:
	print_exit_location();
	return 0;
}

inline int _reuse_fget(long file,long dentry,unsigned long rat_loc)
{
	int err;
	print_entry_location();

	err = _reuse_generic_get(&reuse_rat_fid_file_hash,file,dentry,rat_loc,ARG_FILE);

	print_exit_location();
	return err;
}

inline int _reuse_fput(long file)
{
	int err = 0;
	print_entry_location();

	err = _reuse_generic_put(&reuse_rat_fid_file_hash,file,ARG_FILE);

	print_exit_location();
	return err;
}

inline int _reuse_dget(long dentry,unsigned long inum,unsigned long rat_loc)
{
	int err = 0;
	print_entry_location();

	err = _reuse_generic_get(&reuse_rat_fid_dentry_hash,dentry,inum,rat_loc,ARG_DENTRY);

	print_exit_location();
	return err;
}

inline int _reuse_dput(long dentry)
{
	int err = 0;
	print_entry_location();

	err = _reuse_generic_put(&reuse_rat_fid_dentry_hash,dentry,ARG_DENTRY);

	print_exit_location();
	return err;
}

inline int _reuse_iget(unsigned long inode,unsigned long rat_loc)
{
	int err = 0;
	print_entry_location();

	err = _reuse_generic_get(&reuse_rat_fid_inode_hash,inode,0,rat_loc,ARG_INUM);

	print_exit_location();
	return err;
}

inline int _reuse_iput(unsigned long inode)
{
	int err = 0;
	print_entry_location();

	err = _reuse_generic_put(&reuse_rat_fid_inode_hash,inode,ARG_INUM);

	print_exit_location();
	return err;
}

/*
 * For reusing rat objects (ie. VFS objects).	Once a rat entry is no longer
 * needed, this is called to add to a "free list" and append_rat() will
 * first try to reuse a rat entry from this "free list" before allocating a
 * new rat entry.
 */
int _reuse_rat_obj(int rat_loc)
{
	rat_free_t* entry;
	rat_entry_t* rat;

	print_entry_location();
	assert (rat_loc >= 0);

	entry = rat_free_list_head;
	for(entry = rat_free_list_head;entry != NULL && entry->loc != rat_loc;entry = entry->next);

	if(entry) {
		//Already in free list, don't re-add
		print_warning_message("Rat Location: %d aleady in free list",rat_loc);
		goto out;
	}

	_ADD_TO_LIST(rat_free_list_head,rat_free_list_tail,rat_free_list_tail->next,entry,sizeof(rat_free_t));

	entry->loc = rat_loc;

	//Reset RAT object
	rat = _get_rat(rat_loc);
	if(rat) {
		//Setup the Key
		int keysize = sizeof(rat->rat.u.ival) + sizeof(rat->rat.type);
		char key[keysize];

		memcpy(key,&rat->rat.u.ival,sizeof(rat->rat.u.ival));
		memcpy(key+sizeof(rat->rat.u.ival),&rat->rat.type,sizeof(rat->rat.type));

		rat->rat.type = ARG_INVAL;
		rat->rat.u.ival = 0;

		//Remove from Rat Hash;
		assert(hash_remove(&rat_hash,key,keysize) != NULL);
	}
	else
		print_warning_message("Could not find rat entry: %d to reset", rat_loc);

out:
	print_exit_location();
	return 0;
}

inline int _get_rat_size()
{
	return rat_list_size;
}

static int _destroy_rat()
{
	rat_entry_t* ptr;
	print_entry_location();

	//rat_list is freed during _finalize_prog()
	rat_list_head = NULL;
	rat_list_tail = NULL;
	rat_list_size = 0;

	print_exit_location();
	return 0;
}

int res_manager_init_prog()
{
	print_entry_location();

	prog_list_size = 0;
	prog_list_head = NULL;
	prog_list_tail = NULL;

	print_exit_location();
	return 0;
}

//Inserts rcmdp after prog_entry
int res_manager_insert_prog(prog_entry_t* prog_entry,struct replay_command* rcmdp)
{
	prog_entry_t* entry;
	struct replay_command* srcmdp = NULL;
	print_entry_location();
	assert(prog_entry != NULL);

	//Create and Zero out entry object
	entry = malloc(sizeof(prog_entry_t));
	if(!entry) {
		print_error_nomem();
		exit(ENOMEM);
	}
	memset(entry,0,sizeof(prog_entry_t));

	//Insert into list
	entry->next = prog_entry->next;
	prog_entry->next = entry;
	prog_list_size++;

	//Fill replay command
	srcmdp = malloc(sizeof(struct replay_command));
	if(!srcmdp) {
		print_error_nomem();
		exit(ENOMEM);
	}
	memcpy(srcmdp,rcmdp,sizeof(struct replay_command));
	entry->rcmdp = srcmdp;

	print_exit_location();
	return 0;
}

struct prog_entry_s* res_manager_get_prog_list()
{
	return prog_list_head;
}

struct replay_command* res_manager_append_prog(struct replay_command* rcmdp)
{
	prog_entry_t* entry;
	struct replay_command* srcmdp = NULL;
	print_entry_location();
	assert(filename != NULL);

	_ADD_TO_LIST(prog_list_head,prog_list_tail,prog_list_tail->next,entry,sizeof(prog_entry_t));

	srcmdp = malloc(sizeof(struct replay_command));
	if(!srcmdp) {
		print_error_nomem();
		exit(ENOMEM);
	}
	memcpy(srcmdp,rcmdp,sizeof(struct replay_command));
	entry->rcmdp = srcmdp;
	prog_list_size++;

 out:
	print_exit_location();
	return entry->rcmdp;
}

static int _commit_prog()
{
	int fd = 0;
	int err = 0;
	char* file_tail = ".rep\0";
	char* file_out = malloc(PATH_MAX);
	int sizeof_file_tail = strlen(file_tail);
	prog_entry_t* pptr;
	int curr_cmd = 0;
	print_entry_location();

	if(!file_out) {
		print_error_nomem();
		err = ENOMEM;
		goto out;
	}

	if(sizeof(filename) + sizeof_file_tail > PATH_MAX) {
		print_error("filename is too long\n");
		err = -1;
		goto out;
	}

	strncpy(file_out,filename,PATH_MAX);
	strncat(file_out,file_tail,sizeof_file_tail);

	fd = open(file_out, O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE, 0444);
	if(fd < 0) {
		perror("ERROR: _commit_prog() - Failed during open(2)");
		err = -1;
		goto out;
	}

	//Write Program to Disk and free prog_list
	while(prog_list_head != NULL) {
		assert(prog_list_head->rcmdp != NULL);
		pptr = prog_list_head;
		prog_list_head = prog_list_head->next;

		if(write(fd,pptr->rcmdp,sizeof(struct replay_command)) != sizeof(struct replay_command)) {
			perror("ERROR: _commit_prog() - Failed during write(2)");
			err = -1;
			goto out;
		}
		free(pptr->rcmdp);
		free(pptr);
	 }

	_finalize_prog(fd);
	close(fd);

 out:
	free(file_out);
	if(err < 0)
		exit(err);
	close(fd);
	print_exit_location();
	return err;
}

//Slow implementation; make it faster later by hashing replay_command structs
static prog_entry_t* _find_prog(struct replay_command* rcmdp)
{
	prog_entry_t* pptr = NULL;
	print_entry_location();

	for(pptr = prog_list_head;pptr;pptr = pptr->next)
		if(pptr->rcmdp == rcmdp)
			goto out;

 out:
	print_exit_location();
	return pptr;
}

static int _finalize_prog(int fd)
{
	//Originally finalize_prog() - out.c
	//Mostly the same code with some modifications
	char* buffer;
	int rat_size, size, pages, extra;
	rat_entry_t* ratp;
	rat_entry_t rat_blank;
	int i;

	print_entry_location();
	buffer = (char*)malloc(PAGE_SIZE);
	if (!buffer) {
		print_error_nomem();
		exit(ENOMEM);
	}

	/* here we need to free buffers as soon as they are not needed */
	/* insert operations missing in the trace (open files etc.)*/

	/* insert RAT and headers */
	rat_size = _get_rat_size() * sizeof(struct res_alloc_table);

	size = prog_list_size * sizeof(struct replay_command);
	pages = size/PAGE_SIZE;
	extra = size - pages*PAGE_SIZE;

	if (extra > 0) {
		lseek(fd, pages * PAGE_SIZE, SEEK_SET);
		read(fd, buffer, extra);
		lseek(fd, pages * PAGE_SIZE + rat_size + 24, SEEK_SET);
		write(fd, buffer, extra);
	}
	while (pages > 0) {
		pages--;
		lseek(fd, pages * PAGE_SIZE, SEEK_SET);
		read(fd, buffer, PAGE_SIZE);
		lseek(fd, pages * PAGE_SIZE + rat_size + 24, SEEK_SET);
		write(fd, buffer, PAGE_SIZE);
	}

	lseek(fd, 0, SEEK_SET);

	*((int*)buffer) = *((int*)MAGIC_TRACE);
	write(fd, buffer, 4);
	*((int*)buffer) = TRACE_VER;
	write(fd, buffer, 4);

	*((int*)buffer) = *((int*)MAGIC_RAT);
	write(fd, buffer, 4);
	*((int*)buffer) = rat_size/sizeof(struct res_alloc_table);
	write(fd, buffer, 4);

	//Write blank RATs to disk, reserved for headers
	memset(&rat_blank,0,sizeof(struct res_alloc_table));
	for(i = 0;i < MIN_RAT_SIZE;i++)
		write(fd,(char*)&rat_blank,sizeof(struct res_alloc_table));

	//Write RAT list to disk
	while(rat_list_head != NULL) {
		ratp = rat_list_head;
		rat_list_head = rat_list_head->next;
		write(fd, (char*)&ratp->rat, sizeof(struct res_alloc_table));
		free(ratp);
	}

	*((int*)buffer) = *((int*)MAGIC_CMDS);
	write(fd, buffer, 4);
	*((int*)buffer) = prog_list_size;
	write(fd, buffer, 4);

	free(buffer);

	print_exit_location();
	return 0;
}

static int _destroy_prog()
{
	prog_entry_t* ptr;
	print_entry_location();

	//prog_list is freed during _commit_prog()
	prog_list_head = NULL;
	prog_list_tail = NULL;
	prog_list_size = 0;

	print_exit_location();
	return 0;
}

static int _init_data()
{
	char* file_out = malloc(PATH_MAX);
	char* file_tail = ".rep.data\0";
	int sizeof_tail = strlen(file_tail);
	int err = 0;
	print_entry_location();
	assert(filename != NULL);

	if(!file_out) {
		print_error_nomem();
		err = ENOMEM;
		goto out;
	}

	if(strlen(filename) + sizeof_tail > PATH_MAX) {
		print_error("filename is too long\n");
		err = -1;
		goto out;
	}

	strncpy(file_out,filename,PATH_MAX);
	strncat(file_out,file_tail,PATH_MAX);

	data_fd = open(file_out, O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE, 0444);

	if(data_fd < 0) {
		perror("ERROR: _init_data() - Failed during open(2)");
		err = -1;
		goto out;
	}

	hash_init(&data_list_hash,DATA_LIST_HASH_SIZE);
	if(!ORACLE_MM)
		hash_init(&mm_lru_data_unloaded_hash,MM_LRU_DATA_UNLOADED_HASH_SIZE);
	memory_buffers_size = 0;

 out:
	free(file_out);
	if(err < 0)
		exit(err);
	print_exit_location();
	return 0;
}

int res_manager_append_data(int pointer, timeval_t time, char* buffer, int len,int thread_loc)
{
	int err = 0;
	int rat_loc;
	md5_byte_t hash[MD5_CKSUM_SIZE];
	md5_state_t md5state;
	data_entry_t* entry;
	data_entry_t* ptr;
	print_entry_location();
	assert(filename != NULL);

	//Get the Hash of the buffer
	md5_init(&md5state);
	md5_append(&md5state,(const md5_byte_t *)buffer,len);
	md5_finish(&md5state,hash);

	/*
	 * Determine if the buffer has already been written to disk,
	 * if it's already been written, then "touch" it by moving it to the front
	 * of the LRU list.
	 */
	rat_loc = (int)hash_find(&data_list_hash,(void*)hash,MD5_CKSUM_SIZE);

	if(rat_loc) {
		err = rat_loc;
		_request_reload(hash,data_offset,len,rat_loc,thread_loc);
		goto out;
	}

	//Append to the List
	_ADD_TO_LIST(data_list_head,data_list_tail,data_list_tail->next,entry,sizeof(data_entry_t));

	rat_loc = res_manager_append_rat(pointer,ARG_STR);
	if(rat_loc < 0) {
		print_error("FAILED - Exiting!");
		err = -1;
		goto out;
	}

	/*
	 * Write data and make sure that it is page aligned, this is essential for
	 * replaying without triple buffering.
	 */
	if((data_offset % PAGE_SIZE) + len > PAGE_SIZE) {
		data_offset += PAGE_SIZE - (data_offset % PAGE_SIZE);
		if(lseek(data_fd,data_offset,SEEK_SET) < 0) {
			perror("LSEEK FAILED - ");
			exit(-1);
		}
	}
	assert((data_offset % PAGE_SIZE) + len <= PAGE_SIZE);

	memset(entry,0,sizeof(data_entry_t));
	memcpy(entry->hash,hash,MD5_CKSUM_SIZE);
	entry->len = len;
	entry->data_offset = data_offset;
	entry->time = time;
	entry->rat_loc = rat_loc;
	entry->thread_loc = thread_loc;

	_request_load_op(data_offset,rat_loc,thread_loc,entry);

	//Add to the hash
	assert(!hash_add(&data_list_hash,(void*)&hash,MD5_CKSUM_SIZE,(void*)rat_loc));

	if(write(data_fd,buffer,len) != len) {
		perror("ERROR: res_manager_append_data() - Failed during write(2)");
		err = -1;
		goto out;
	}

	err = rat_loc;
	data_offset += len;

out:
	if(err < 0)
		exit(err);
	print_exit_location();
	return err;
}

static int _destroy_data()
{
	data_entry_t* dptr;
	print_entry_location();

	while(data_list_head != NULL) {
		dptr = data_list_head;
		data_list_head = data_list_head->next;
		free(dptr);
	}

	data_list_head = NULL;
	data_list_tail = NULL;

	hash_destroy(&data_list_hash);
	hash_destroy(&mm_lru_data_unloaded_hash);

	if(close(data_fd)) {
		perror("ERROR: _destroy_data() - Failed on close(2)");
		exit(-1);
	}

	print_exit_location();
	return 0;
}

static int _optimize()
{
	print_entry_location();

	_optimize_rat();

	print_exit_location();
	return 0;
}

//Optimizes the RAT's memory usage.
static int _optimize_rat()
{
	print_entry_location();

	//_add_unload_ops_to_last_mem_ref(); //Add unload ops for memory buffers

	print_exit_location();
	return 0;
}

//For Buffers - unload buffers after last use
static int _add_unload_ops_to_last_mem_ref()
{
	rat_ref_t* rrptr;
	prog_entry_t* prog_entry;
	prog_entry_t* pptr;
	struct replay_command* unload_cmd;
	print_entry_location();

	for(rrptr = rat_ref_list_head;rat_ref_list_head;rrptr = rat_ref_list_head) {
		rat_ref_list_head = rat_ref_list_head->next;
		//Set up unload_command
		unload_cmd = malloc(sizeof(struct replay_command));
		if(!unload_cmd) {
			print_error_nomem();
			exit(ENOMEM);
		}

		memset(unload_cmd,0,sizeof(struct replay_command));
		unload_cmd->command = OP_UNLOAD;
		unload_cmd->ret = 0;
		unload_cmd->args[0] = rrptr->rat_loc;
		unload_cmd->time = rrptr->rcmdp->time;

		//Manually Insert into prog_list
		prog_entry = malloc(sizeof(prog_entry_t));
		if(!prog_entry) {
			print_error_nomem();
			exit(ENOMEM);
		}

		pptr = _find_prog(rrptr->rcmdp);
		assert(pptr != NULL);
		prog_entry->rcmdp = unload_cmd;
		prog_entry->next = pptr->next;
		pptr->next = prog_entry;
		prog_list_size++;

		free(rrptr);
	}

out:
	print_exit_location();
	return 0;
}
